# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import numpy as np

# lon1, lat1, lon2, lat2 in degrees
coords = [(1, 0, 0, 0),
          (0, 1, 0, 0),
          (0, 0, 1, 0),
          (0, 0, 0, 1),
          (0, 0, 10, 0),
          (0, 0, 90, 0),
          (0, 0, 180, 0),
          (0, 45, 0, -45),
          (0, 60, 0, -30),
          (-135, -15, 45, 15),
          (100, -89, -80, 89),
          (0, 0, 0, 0),
          (0, 0, 1. / 60., 1. / 60.)]
correct_seps = [1, 1, 1, 1, 10, 90, 180, 90, 90, 180, 180, 0,
                0.023570225877234643]
correctness_margin = 2e-10


def test_fk5_seps():
    """
    This tests if `separation` works for FK5Coordinate objects.

    This is a regression test for github issue #891
    """
    from astropy.coordinates import FK5

    a = FK5(1., 1., unit=('deg', 'deg'))
    b = FK5(2., 2., unit=('deg', 'deg'))
    a.separation(b)


def test_angsep():
    """
    Tests that the angular separation object also behaves correctly.
    """

    from ...units import Quantity
    from ..angles import Angle
    from ..angle_utilities import angular_separation

    # check it both works with floats in radians, Quantities, or Angles
    for conv in (np.deg2rad,
                 lambda x: Quantity(x, "deg"),
                 lambda x: Angle(x, "deg")):
        for (lon1, lat1, lon2, lat2), corrsep in zip(coords, correct_seps):
            angsep = angular_separation(conv(lon1), conv(lat1),
                                        conv(lon2), conv(lat2))
            assert np.fabs(angsep - conv(corrsep)) < conv(correctness_margin)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test `astropy.utils.timer`.

.. note::

    The tests only compare rough estimates as
    performance is machine-dependent.

"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

# STDLIB
import time

# LOCAL
from ..timer import RunTimePredictor


def func_to_time(x):
    """This is sleeps for x seconds for timing tests."""
    time.sleep(x)
    return 'Slept for {0} second(s)'.format(x)


def test_timer():
    p = RunTimePredictor(func_to_time)
    try:
        p.do_fit()
    except AssertionError as e:
        assert str(e) == 'Requires 3 points but has 0'

    try:
        p.predict_time(100)
    except AssertionError as e:
        assert str(e) == 'No fitted data for prediction'

    p.time_func([0.1, 0.2, 0.5, 'a', 1.5])
    p.time_func(1.0)

    assert p._funcname == 'func_to_time'
    assert p._cache_bad == ['a']
    assert p.results == {0.1: 'Slept for 0.1 second(s)',
                         0.2: 'Slept for 0.2 second(s)',
                         0.5: 'Slept for 0.5 second(s)',
                         1.5: 'Slept for 1.5 second(s)',
                         1.0: 'Slept for 1.0 second(s)'}

    a = p.do_fit()
    assert p._power == 1

    # Perfect slope is 1, with 10% uncertainty
    assert 0.9 <= a[0] <= 1.1

    # Perfect intercept is 0, with 1-sec uncertainty
    assert -1 <= a[1] <= 1

    # Perfect answer is 100, with 10% uncertainty
    t = p.predict_time(100)
    assert 90 <= t <= 110

    # Repeated call to access cached run time
    t2 = p.predict_time(100)
    assert t == t2



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import itertools

import numpy as np
from numpy.testing import assert_allclose

from ...tests.helper import pytest

from ..utils import discretize_model
from ...modeling.functional_models import (
    Gaussian1D, Box1D, MexicanHat1D, Trapezoid1D,
    Gaussian2D, Box2D, MexicanHat2D)
from ...modeling.tests.example_models import models_1D, models_2D
from ...modeling.tests.test_models import create_model

try:
    import scipy
    HAS_SCIPY = True
except ImportError:
    HAS_SCIPY = False


modes = ['center', 'linear_interp', 'oversample']
test_models_1D = [Gaussian1D, Box1D, MexicanHat1D]
test_models_2D = [Gaussian2D, Box2D, MexicanHat2D]


@pytest.mark.parametrize(('model_class', 'mode'), list(itertools.product(test_models_1D, modes)))
def test_pixel_sum_1D(model_class, mode):
    """
    Test if the sum of all pixels corresponds nearly to the integral.
    """
    if model_class == Box1D and mode == "center":
        pytest.skip("Non integrating mode. Skip integral test.")
    parameters = models_1D[model_class]['parameters']
    model = create_model(model_class, parameters)

    values = discretize_model(model, models_1D[model_class]['x_lim'], mode=mode)
    assert_allclose(values.sum(), models_1D[model_class]['integral'], atol=0.0001)


@pytest.mark.parametrize('mode', modes)
def test_gaussian_eval_1D(mode):
    """
    Discretize Gaussian with different modes and check
    if result is at least similar to Gaussian1D.eval().
    """
    model = Gaussian1D(1, 0, 20)
    x = np.arange(-100, 101)
    values = model(x)
    disc_values = discretize_model(model, (-100, 101), mode=mode)
    assert_allclose(values, disc_values, atol=0.001)


@pytest.mark.parametrize(('model_class', 'mode'), list(itertools.product(test_models_2D, modes)))
def test_pixel_sum_2D(model_class, mode):
    """
    Test if the sum of all pixels corresponds nearly to the integral.
    """
    if model_class == Box2D and mode == "center":
        pytest.skip("Non integrating mode. Skip integral test.")

    parameters = models_2D[model_class]['parameters']
    model = create_model(model_class, parameters)

    values = discretize_model(model, models_2D[model_class]['x_lim'],
                              models_2D[model_class]['y_lim'], mode=mode)
    assert_allclose(values.sum(), models_2D[model_class]['integral'], atol=0.0001)


@pytest.mark.parametrize('mode', modes)
def test_gaussian_eval_2D(mode):
    """
    Discretize Gaussian with different modes and check
    if result is at least similar to Gaussian1D.eval()
    """
    model = Gaussian2D(1, 0, 0, 20, 20)
    x = np.arange(-100, 101)
    y = np.arange(-100, 101)
    y, x = np.meshgrid(y, x)
    values = model(x, y)
    disc_values = discretize_model(model, (-100, 101), (-100, 101), mode=mode)
    assert_allclose(values, disc_values, atol=0.001)


@pytest.mark.skipif('not HAS_SCIPY')
def test_subpixel_gauss_1D():
    """
    Test subpixel accuracy of the oversample mode with gaussian 1D model.
    """
    gauss_1D = Gaussian1D(1, 0, 0.1)
    values = discretize_model(gauss_1D, (-1, 2), mode='integrate', factor=100)
    assert_allclose(values.sum(), np.sqrt(2 * np.pi) * 0.1, atol=0.00001)


@pytest.mark.skipif('not HAS_SCIPY')
def test_subpixel_gauss_2D():
    """
    Test subpixel accuracy of the oversample mode with gaussian 2D model.
    """
    gauss_2D = Gaussian2D(1, 0, 0, 0.1, 0.1)
    values = discretize_model(gauss_2D, (-1, 2), (-1, 2), mode='integrate', factor=100)
    assert_allclose(values.sum(), 2 * np.pi * 0.01, atol=0.00001)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import absolute_import, division, print_function, unicode_literals

import abc

import numpy as np

__all__ = ['MissingDataAssociationException',
           'IncompatibleUncertaintiesException', 'NDUncertainty',
           'StdDevUncertainty']


class IncompatibleUncertaintiesException(Exception):
    """
    This exception should be used to indicate cases in which uncertainties
    with two different classes can not be propagated.
    """
    pass


class MissingDataAssociationException(Exception):
    """
    This exception should be used to indicate that an uncertainty instance has
    not been associated with a parent `~astropy.nddata.nddata.NDData` object.
    """
    pass


class NDUncertainty(object):
    '''
    This is the base class for uncertainty classes used with NDData. It is
    implemented as an abstract class and should never be directly
    instantiated.

    Classes inheriting from NDData should overload the ``propagate_*``
    methods, keeping the call signature the same. The propagate methods can
    assume that a `parent_nddata` attribute is present which links to the parent_nddata
    dataset, and take an `~astropy.nddata.NDData` instance as the positional
    argument, *not* an `~astropy.nddata.NDUncertainty` instance, because the
    `~astropy.nddata.NDData` instance can be used to access both the data and
    the uncertainties (some propagations require the data values).
    '''

    __metaclass__ = abc.ABCMeta

    # Indicates whether the class supports the propagation of correlated
    # uncertainties
    supports_correlated = False

    @property
    def parent_nddata(self):
        if self._parent_nddata is None:
            raise MissingDataAssociationException("uncertainty is not associated with an NDData object")
        else:
            return self._parent_nddata

    @parent_nddata.setter
    def parent_nddata(self, value):
        self._parent_nddata = value

    @abc.abstractmethod
    def propagate_add(self, other_nddata, result_data):
        '''
        Propagate uncertainties for addition.

        Parameters
        ----------
        other_nddata : NDData instance
            The data for the second other_nddata in a + b
        result_data : `~numpy.ndarray` instance
            The data array that is the result of the addition

        Returns
        -------
        result_uncertainty : NDUncertainty instance
            The resulting uncertainty

        Raises
        ------
        IncompatibleUncertaintiesException
            Raised if the method does not know how to add the uncertainties
        '''
        pass

    @abc.abstractmethod
    def propagate_subtract(self, other_nddata, result_data):
        '''
        Propagate uncertainties for subtraction.

        Parameters
        ----------
        other_nddata : NDData instance
            The data for the second other_nddata in a + b
        result_data : `~numpy.ndarray` instance
            The data array that is the result of the addition

        Returns
        -------
        result_uncertainty : NDUncertainty instance
            The resulting uncertainty

        Raises
        ------
        IncompatibleUncertaintiesException
            Raised if the method does not know how to add the uncertainties
        '''
        pass

    @abc.abstractmethod
    def propagate_multiply(self, other_nddata, result_data):
        '''
        Propagate uncertainties for multiplication.

        Parameters
        ----------
        other_nddata : NDData instance
            The data for the second other_nddata in a + b
        result_data : `~numpy.ndarray` instance
            The data array that is the result of the addition

        Returns
        -------
        result_uncertainty : NDUncertainty instance
            The resulting uncertainty
        '''
        pass

    @abc.abstractmethod
    def propagate_divide(self, other_nddata, result_data):
        '''
        Propagate uncertainties for division.

        Parameters
        ----------
        other_nddata : NDData instance
            The data for the second other_nddata in a + b
        result_data : `~numpy.ndarray` instance
            The data array that is the result of the addition

        Returns
        -------
        result_uncertainty : NDUncertainty instance
            The resulting uncertainty
        '''
        pass


class StdDevUncertainty(NDUncertainty):
    '''
    A class for standard deviation uncertaintys
    '''

    support_correlated = False

    def __init__(self, array=None, copy=True):
        if array is None:
            self.array = None
        elif isinstance(array, StdDevUncertainty):
            self.array = np.array(array.array, copy=copy, subok=True)
        else:
            self.array = np.array(array, copy=copy, subok=True)

    @property
    def parent_nddata(self):
        try:
            if self._parent_nddata is None:
                raise MissingDataAssociationException("Uncertainty is not associated with an NDData object")
            else:
                return self._parent_nddata
        except AttributeError:
            raise MissingDataAssociationException("Uncertainty is not associated with an NDData object")

    @parent_nddata.setter
    def parent_nddata(self, value):
        if self.array is None or value is None:
            self._parent_nddata = value
        else:
            if value.shape != self.array.shape:
                raise ValueError("parent shape does not match array data shape")
        self._parent_nddata = value

    @property
    def array(self):
        return self._array

    @array.setter
    def array(self, value):
        if value is not None:
            try:
                if value.shape != self.parent_nddata.shape:
                    raise ValueError("array shape does not match parent data shape")
            except MissingDataAssociationException:
                pass
        self._array = value

    def propagate_add(self, other_nddata, result_data):
        '''
        Propagate uncertainties for addition.

        Parameters
        ----------
        other_nddata : NDData instance
            The data for the second other_nddata in a + b
        result_data : `~numpy.ndarray` instance
            The data array that is the result of the addition

        Returns
        -------
        result_uncertainty : NDUncertainty instance
            The resulting uncertainty

        Raises
        ------
        IncompatibleUncertaintiesException
            Raised if the method does not know how to propagate the uncertainties
        '''

        if not isinstance(other_nddata.uncertainty, StdDevUncertainty):
            raise IncompatibleUncertaintiesException

        if self.array is None:
            raise ValueError("standard deviation values are not set")

        if other_nddata.uncertainty.array is None:
            raise ValueError("standard deviation values are not set in other_nddata")

        result_uncertainty = StdDevUncertainty()
        result_uncertainty.array = np.sqrt(self.array ** 2 + other_nddata.uncertainty.array ** 2)

        return result_uncertainty

    def __getitem__(self, item):
        '''
            Slice the standard deviation array using standard numpy slicing
        '''

        new_array = self.array[item]
        return self.__class__(new_array, copy=False)

    def propagate_subtract(self, other_nddata, result_data):
        '''
        Propagate uncertainties for subtraction.

        Parameters
        ----------
        other_nddata : NDData instance
            The data for the second other_nddata in a + b
        result_data : `~numpy.ndarray` instance
            The data array that is the result of the addition

        Returns
        -------
        result_uncertainty : NDUncertainty instance
            The resulting uncertainty

        Raises
        ------
        IncompatibleUncertaintiesException
            Raised if the method does not know how to propagate the uncertainties
        '''

        if not isinstance(other_nddata.uncertainty, StdDevUncertainty):
            raise IncompatibleUncertaintiesException

        if self.array is None:
            raise ValueError("standard deviation values are not set")

        if other_nddata.uncertainty.array is None:
            raise ValueError("standard deviation values are not set in other_nddata")

        result_uncertainty = StdDevUncertainty()
        result_uncertainty.array = np.sqrt(self.array ** 2 + other_nddata.uncertainty.array ** 2)

        return result_uncertainty

    def propagate_multiply(self, other_nddata, result_data):
        '''
        Propagate uncertainties for mutliplication.

        Parameters
        ----------
        other_nddata : NDData instance
            The data for the second other_nddata in a + b
        result_data : `~numpy.ndarray` instance
            The data array that is the result of the addition

        Returns
        -------
        result_uncertainty : NDUncertainty instance
            The resulting uncertainty

        Raises
        ------
        IncompatibleUncertaintiesException
            Raised if the method does not know how to propagate the uncertainties
        '''

        if not isinstance(other_nddata.uncertainty, StdDevUncertainty):
            raise IncompatibleUncertaintiesException

        if self.array is None:
            raise ValueError("standard deviation values are not set")

        if other_nddata.uncertainty.array is None:
            raise ValueError("standard deviation values are not set in other_nddata")

        result_uncertainty = StdDevUncertainty()
        result_uncertainty.array = (np.sqrt((self.array / self.parent_nddata.data) ** 2
                                            + (other_nddata.uncertainty.array / other_nddata.data) ** 2)
                                    * result_data)

        return result_uncertainty

    def propagate_divide(self, other_nddata, result_data):
        '''
        Propagate uncertainties for division.

        Parameters
        ----------
        other_nddata : NDData instance
            The data for the second other_nddata in a + b
        result_data : `~numpy.ndarray` instance
            The data array that is the result of the addition

        Returns
        -------
        result_uncertainty : NDUncertainty instance
            The resulting uncertainty

        Raises
        ------
        IncompatibleUncertaintiesException
            Raised if the method does not know how to propagate the uncertainties
        '''

        if not isinstance(other_nddata.uncertainty, StdDevUncertainty):
            raise IncompatibleUncertaintiesException

        if self.array is None:
            raise ValueError("standard deviation values are not set")

        if other_nddata.uncertainty.array is None:
            raise ValueError("standard deviation values are not set in other_nddata")

        result_uncertainty = StdDevUncertainty()
        result_uncertainty.array = (np.sqrt((self.array / self.parent_nddata.data) ** 2
                                            + (other_nddata.uncertainty.array / other_nddata.data) ** 2)
                                    * result_data)

        return result_uncertainty



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Common utilities for accessing VO simple services."""
from __future__ import print_function, division

# STDLIB
import fnmatch
import json
import re
import urllib

# LOCAL
from ...config.configuration import ConfigurationItem
from ...io.votable import table, tree
from ...io.votable.exceptions import vo_raise, vo_warn, E19, W24, W25
from ...utils.console import color_print
from ...utils.data import get_readable_fileobj


__all__ = ['VOSCatalog', 'VOSDatabase', 'get_remote_catalog_db',
           'call_vo_service', 'list_catalogs']

__dbversion__ = 1

VO_PEDANTIC = table.PEDANTIC()

BASEURL = ConfigurationItem('vos_baseurl',
                            'http://stsdas.stsci.edu/astrolib/vo_databases/',
                            'URL where VO Service database file is stored.')


class VOSError(Exception):  # pragma: no cover
    pass


class VOSCatalog(object):
    """A class to represent VO Service Catalog.

    Parameters
    ----------
    tree : JSON tree

    """
    def __init__(self, tree):
        self._tree = tree

    def __getattr__(self, what):
        """Expose dictionary attributes."""
        return getattr(self._tree, what)

    def __getitem__(self, what):
        """Expose dictionary key look-up."""
        return self._tree[what]

    def __str__(self):  # pragma: no cover
        """Show the most important and unique things about a catalog."""
        keys = ('title', 'url')
        out_str = '\n'.join(['{0}: {1}'.format(key, self._tree[key])
                             for key in keys if key in self._tree])
        return out_str

    def dumps(self):  # pragma: no cover
        """Dump the contents into a string.

        Returns
        -------
        s : str
            Contents as JSON string dump.

        """
        return json.dumps(self._tree, sort_keys=True, indent=4)


class VOSDatabase(VOSCatalog):
    """A class to represent a collection of `VOSCatalog`.

    Parameters
    ----------
    tree : JSON tree

    Raises
    ------
    VOSError
        If given ``tree`` does not have 'catalogs' key.

    """
    def __init__(self, tree):
        self._tree = tree

        if tree['__version__'] > __dbversion__:  # pragma: no cover
            vo_warn(W24)

        if not 'catalogs' in tree:  # pragma: no cover
            raise VOSError("Invalid VO service catalog database")

        self._catalogs = tree['catalogs']

    def __str__(self):  # pragma: no cover
        """Show the most important and unique things about a database."""
        return '\n'.join(sorted(self._catalogs.keys()))

    def get_catalogs(self):
        """Iterator to get all catalogs."""
        for key, val in self._catalogs.items():
            yield key, VOSCatalog(val)

    def get_catalogs_by_url(self, url):
        """Like :func:`get_catalogs` but using access URL look-up."""
        for key, cat in self.get_catalogs():
            if cat['url'] == url:
                yield key, cat

    def get_catalog(self, name):
        """Get one catalog of given name.

        Parameters
        ----------
        name : str
            Primary key identifying the catalog.

        Returns
        -------
        obj : `VOSCatalog` object

        Raises
        ------
        VOSError
            If catalog is not found.

        """
        if not name in self._catalogs:
            raise VOSError("No catalog '{0}' found.".format(name))
        return VOSCatalog(self._catalogs[name])

    def get_catalog_by_url(self, url):
        """Like :func:`get_catalog` but using access URL look-up.
        On multiple matches, only first match is returned.

        """
        out_cat = None
        for key, cat in self.get_catalogs_by_url(url):
            out_cat = cat
            break
        if out_cat is None:  # pragma: no cover
            raise VOSError("No catalog with URL '{0}' found.".format(url))
        return out_cat

    def list_catalogs(self, pattern=None, sort=True):
        """List catalog names.

        Parameters
        ----------
        pattern : str or `None`
            If given string is anywhere in a catalog name, it is
            considered a matching catalog. It accepts patterns as
            in :py:mod:`fnmatch` and is case-insensitive.
            By default, all catalogs are returned.

        sort : bool
            Sort output in alphabetical order. If not sorted, the
            order depends on dictionary hashing. Default is `True`.

        Returns
        -------
        out_arr : list of str
            List of catalog names.

        """
        all_catalogs = self._catalogs.keys()

        if pattern is None or len(all_catalogs) == 0:
            out_arr = all_catalogs
        else:
            pattern = re.compile(fnmatch.translate('*' + pattern + '*'),
                                 re.IGNORECASE)
            out_arr = [s for s in all_catalogs if pattern.match(s)]

        if sort:
            out_arr.sort()

        return out_arr


def get_remote_catalog_db(dbname, cache=True, verbose=True):
    """Get a database of VO services (which is a JSON file) from a remote
    location.

    Parameters
    ----------
    dbname : str
        Prefix of JSON file to download from
        ``astropy.vo.client.vos_catalog.BASEURL``.

    cache : bool
        Use caching for VO Service database. Access to actual VO
        websites referenced by the database still needs internet
        connection.

    verbose : bool
        Show download progress bars.

    Returns
    -------
    obj : `VOSDatabase` object
        A database of VO services.

    """
    with get_readable_fileobj(BASEURL() + dbname + '.json',
                              encoding='utf8', cache=cache,
                              show_progress=verbose) as fd:
        tree = json.load(fd)

    return VOSDatabase(tree)


def _get_catalogs(service_type, catalog_db, **kwargs):
    """Expand ``catalog_db`` to a list of catalogs.

    Parameters
    ----------
    service_type, catalog_db
        See :func:`call_vo_service`.

    kwargs : dict
        Keywords accepted by :func:`get_remote_catalog_db`.

    Returns
    -------
    catalogs : list of tuple
        List of catalogs in the form of ``(key, VOSCatalog)``.

    """
    if catalog_db is None:
        catalog_db = get_remote_catalog_db(service_type, **kwargs)
        catalogs = catalog_db.get_catalogs()
    elif isinstance(catalog_db, VOSDatabase):
        catalogs = catalog_db.get_catalogs()
    elif isinstance(catalog_db, (VOSCatalog, basestring)):
        catalogs = [(None, catalog_db)]
    elif isinstance(catalog_db, list):
        for x in catalog_db:
            assert (isinstance(x, (VOSCatalog, basestring)) and
                    not isinstance(x, VOSDatabase))
        catalogs = [(None, x) for x in catalog_db]
    else:  # pragma: no cover
        raise VOSError('catalog_db must be a catalog database, '
                       'a list of catalogs, or a catalog')

    return catalogs


def _vo_service_request(url, pedantic, kwargs, verbose=False):
    if len(kwargs) and not (url.endswith('?') or url.endswith('&')):
        raise VOSError("url should already end with '?' or '&'")

    query = []
    for key, value in kwargs.iteritems():
        query.append('{}={}'.format(
            urllib.quote(key), urllib.quote_plus(str(value))))

    parsed_url = url + '&'.join(query)
    with get_readable_fileobj(parsed_url, encoding='binary',
                              show_progress=verbose) as req:
        tab = table.parse(req, filename=parsed_url, pedantic=pedantic)

    return vo_tab_parse(tab, url, kwargs)


def vo_tab_parse(tab, url, kwargs):
    """In case of errors from the server, a complete and correct
    'stub' VOTable file may still be returned.  This is to
    detect that case.

    Parameters
    ----------
    tab : `astropy.io.votable.tree.VOTableFile` object

    url : str
        URL used to obtain ``tab``.

    kwargs : dict
        Keywords used to obtain ``tab``, if any.

    Returns
    -------
    out_tab : `astropy.io.votable.tree.Table` object

    Raises
    ------
    IndexError
        Table iterator fails.

    VOSError
        Server returns error message or invalid table.

    """
    for param in tab.iter_fields_and_params():
        if param.ID is not None and param.ID.lower() == 'error':
            if isinstance(param, tree.Param):
                e = param.value
            else:  # pragma: no cover
                e = ''
            raise VOSError("Catalog server '{0}' returned error '{1}'".format(
                url, e))

    for info in tab.infos:
        if info.name is not None and info.name.lower() == 'error':
            raise VOSError("Catalog server '{0}' returned error '{1}'".format(
                url, info.value))

    if tab.resources == []:  # pragma: no cover
        vo_raise(E19)

    for info in tab.resources[0].infos:
        if ((info.name == 'QUERY_STATUS' and info.value != 'OK') or
                (info.name is not None and info.name.lower() == 'error')):
            if info.content is not None:  # pragma: no cover
                long_descr = ':\n{0}'.format(info.content)
            else:
                long_descr = ''
            raise VOSError("Catalog server '{0}' returned status "
                           "'{1}'{2}".format(url, info.value, long_descr))

    out_tab = tab.get_first_table()

    kw_sr = [k for k in kwargs if 'sr' == k.lower()]
    if len(kw_sr) == 0:
        sr = 0
    else:
        sr = kwargs.get(kw_sr[0])

    if sr != 0 and out_tab.array.size <= 0:
        raise VOSError("Catalog server '{0}' returned {1} result".format(
            url, out_tab.array.size))

    out_tab.url = url  # Track the URL
    return out_tab


def call_vo_service(service_type, catalog_db=None, pedantic=None,
                    verbose=True, cache=True, kwargs={}):
    """Makes a generic VO service call.

    Parameters
    ----------
    service_type : str
        Name of the type of service, e.g., 'conesearch_good'.
        Used in error messages and to select a catalog database
        if ``catalog_db`` is not provided.

    catalog_db
        May be one of the following, in order from easiest to
        use to most control:

            - `None`: A database of ``service_type`` catalogs is downloaded
              from ``astropy.vo.client.vos_catalog.BASEURL``.  The first
              catalog in the database to successfully return a result is used.

            - *catalog name*: A name in the database of ``service_type``
              catalogs at ``astropy.vo.client.vos_catalog.BASEURL`` is used.
              For a list of acceptable names, use :func:`list_catalogs`.

            - *url*: The prefix of a URL to a IVOA Service for
              ``service_type``. Must end in either '?' or '&'.

            - `VOSCatalog` object: A specific catalog manually downloaded and
              selected from the database (see :ref:`vo-sec-client-vos`).

            - Any of the above 3 options combined in a list, in which case
              they are tried in order.

    pedantic : bool or `None`
        When `True`, raise an error when the file violates the spec,
        otherwise issue a warning.  Warnings may be controlled using
        :py:mod:`warnings` module.
        When not provided, uses the configuration setting
        ``astropy.io.votable.table.PEDANTIC``, which defaults to `False`.

    verbose : bool
        Verbose output.

    cache : bool
        Use caching for VO Service database. Access to actual VO
        websites referenced by the database still needs internet
        connection.

    kwargs : dictionary
        Keyword arguments to pass to the catalog service.
        No checking is done that the arguments are accepted by
        the service, etc.

    Returns
    -------
    obj : `astropy.io.votable.tree.Table` object
        First table from first successful VO service request.

    Raises
    ------
    VOSError
        If VO service request fails.

    """
    catalogs = _get_catalogs(service_type, catalog_db, cache=cache,
                             verbose=verbose)

    if pedantic is None:  # pragma: no cover
        pedantic = VO_PEDANTIC

    for name, catalog in catalogs:
        if isinstance(catalog, basestring):
            if catalog.startswith("http"):
                url = catalog
            else:
                remote_db = get_remote_catalog_db(service_type, cache=cache,
                                                  verbose=verbose)
                catalog = remote_db.get_catalog(catalog)
                url = catalog['url']
        else:
            url = catalog['url']

        if verbose:  # pragma: no cover
            color_print('Trying {0}'.format(url), 'green')

        try:
            return _vo_service_request(url, pedantic, kwargs, verbose=verbose)
        except Exception as e:
            vo_warn(W25, (url, str(e)))

    raise VOSError('None of the available catalogs returned valid results.')


def list_catalogs(service_type, cache=True, verbose=True, **kwargs):
    """List the catalogs available for the given service type.

    Parameters
    ----------
    service_type : str
        Name of the type of service, e.g., 'conesearch_good'.

    cache : bool
        Use caching for VO Service database. Access to actual VO
        websites referenced by the database still needs internet
        connection.

    verbose : bool
        Show download progress bars.

    pattern : str or `None`
        If given string is anywhere in a catalog name, it is
        considered a matching catalog. It accepts patterns as
        in :py:mod:`fnmatch` and is case-insensitive.
        By default, all catalogs are returned.

    sort : bool
        Sort output in alphabetical order. If not sorted, the
        order depends on dictionary hashing. Default is `True`.

    Returns
    -------
    arr : list of str
        List of catalog names.

    """
    return get_remote_catalog_db(service_type, cache=cache,
                                 verbose=verbose).list_catalogs(**kwargs)



import warnings

from ..utils.exceptions import AstropyUserWarning

import numpy as np
cimport numpy as np
import cython

ctypedef np.double_t DOUBLE_T

cdef extern from "erfa.h":
    double eraEpb(double dj1, double dj2)
    double eraEpj(double dj1, double dj2)
    void eraEpj2jd(double epj, double *djm0, double *djm)
    void eraEpb2jd(double epb, double *djm0, double *djm)
    int eraCal2jd(int iy, int im, int id, double *djm0, double *djm)
    int eraJd2cal(double dj1, double dj2,int *iy, int *im, int *id, double *fd)
    int eraJdcalf(int ndp, double dj1, double dj2, int iymdf[4])

    # Internal JD to/from datetime format converters
    int eraD2dtf(char *scale, int ndp, double d1, double d2, int *iy, int *im, int *id, int ihmsf[4])
    int eraDtf2d(char *scale, int iy, int im, int id, int ihr, int imn, double sec, double *d1, double *d2)

    # Time scale helper routines
    double eraDtdb(double date1, double date2, double ut, double elong, double u, double v)
    int eraDat(int iy, int im, int id, double fd, double *deltat)

    # Time scale conversion routines
    int eraTaitt(double tai1, double tai2, double *tt1, double *tt2)
    int eraTttai(double tt1, double tt2, double *tai1, double *tai2)
    int eraTaiutc(double tai1, double tai2, double *utc1, double *utc2)
    int eraUtctai(double utc1, double utc2, double *tai1, double *tai2)
    int eraTcbtdb(double tcb1, double tcb2, double *tdb1, double *tdb2)
    int eraTdbtcb(double tdb1, double tdb2, double *tcb1, double *tcb2)
    int eraTcgtt(double tcg1, double tcg2, double *tt1, double *tt2)
    int eraTttcg(double tt1, double tt2, double *tcg1, double *tcg2)

    int eraTaiut1(double tai1, double tai2, double dta, double *ut11, double *ut12)
    int eraUt1tai(double ut11, double ut12, double dta, double *tai1, double *tai2)
    int eraTtut1(double tt1, double tt2, double dt, double *ut11, double *ut12)
    int eraUt1tt(double ut11, double ut12, double dt, double *tt1, double *tt2)
    int eraTdbtt(double tdb1, double tdb2, double dtr, double *tt1, double *tt2)
    int eraTttdb(double tt1, double tt2, double dtr, double *tdb1, double *tdb2)
    int eraUt1utc(double ut11, double ut12, double dut1, double *utc1, double *utc2)
    int eraUtcut1(double utc1, double utc2, double dut1, double *ut11, double *ut12)

    # Geodetic
    int eraAf2a(char s, int ideg, int iamin, double asec, double *rad)
    int eraGd2gc(int n, double elong, double phi, double height, double xyz[3])

DUBIOUS = 'dubious year for UTC (before 1960.0 or 5 years ' \
          'beyond last known leap second)'

def check_return(ret, func_name, warns={}, errors={}):
    """Check the return value from an era routine"""
    if ret in warns:
        warnings.warn('{0}: {1}'.format(func_name, warns[ret]), AstropyUserWarning)
    elif ret in errors:
        raise ValueError('{0}: {1}'.format(func_name, errors[ret]))
    elif ret != 0:
        raise ValueError('Unexpected return code {0} from {1}'
                         .format(repr(ret), func_name))


@cython.wraparound(False)
@cython.boundscheck(False)
def cal2jd( 
    np.ndarray[int, ndim=1] iy,
    np.ndarray[int, ndim=1] im,
    np.ndarray[int, ndim=1] id,
    np.ndarray[double, ndim=1] djm0,
    np.ndarray[double, ndim=1] djm):
    """
    int eraCal2jd(int iy, int im, int id, double *djm0, double *djm)
    Calendar date to high-precision JD.

    **  Given:
    **     iy,im,id  int     year, month, day in Gregorian calendar (Note 1)
    **
    **  Returned:
    **     djm0      double  MJD zero-point: always 2400000.5
    **     djm       double  Modified Julian Date for 0 hrs
    **
    **  Returned (function value):
    **               int     status:
    **                           0 = OK
    **                          -1 = bad year   (Note 3: JD not computed)
    **                          -2 = bad month  (JD not computed)
    **                          -3 = bad day    (JD computed)
    **
    **  Notes:
    **
    **  1) The algorithm used is valid from -4800 March 1, but this
    **     implementation rejects dates before -4799 January 1.
    **
    **  2) The Julian Date is returned in two pieces, in the usual ERFA
    **     manner, which is designed to preserve time resolution.  The
    **     Julian Date is available as a single number by adding djm0 and
    **     djm.
    **
    **  3) In early eras the conversion is from the "Proleptic Gregorian
    **     Calendar";  no account is taken of the date(s) of adoption of
    **     the Gregorian Calendar, nor is the AD/BC numbering convention
    **     observed.
    """
    cdef unsigned int i
    cdef unsigned n = iy.shape[0]
    warns = {-3: 'Bad input day (JD still computed)'}
    errs = {-1: 'Bad input year',
             -2: 'Bad input month'}
    for i in range(n):
        ret = eraCal2jd( iy[i], im[i], id[i], &djm0[i], &djm[i])
        check_return(ret, 'eraCal2jd', warns, errs)
    return

@cython.wraparound(False)
@cython.boundscheck(False)
def d_tai_utc(np.ndarray[int, ndim=1] iy,
              np.ndarray[int, ndim=1] im,
              np.ndarray[int, ndim=1] id,
              np.ndarray[double, ndim=1] fd):
    """
    int eraDat(int iy, int im, int id, double fd, double *deltat)
    For a given UTC date, calculate delta(AT) = TAI-UTC.

    **  Given:
    **     iy     int      UTC:  year (Notes 1 and 2)
    **     im     int            month (Note 2)
    **     id     int            day (Notes 2 and 3)
    **     fd     double         fraction of day (Note 4)
    **
    **  Returned:
    **     deltat double   TAI minus UTC, seconds
    **
    **  Returned (function value):
    **            int      status (Note 5):
    **                       1 = dubious year (Note 1)
    **                       0 = OK
    **                      -1 = bad year
    **                      -2 = bad month
    **                      -3 = bad day (Note 3)
    **                      -4 = bad fraction (Note 4)
    **
    **  Notes:
    **
    **  1) UTC began at 1960 January 1.0 (JD 2436934.5) and it is improper
    **     to call the function with an earlier date.  If this is attempted,
    **     zero is returned together with a warning status.
    **
    **     Because leap seconds cannot, in principle, be predicted in
    **     advance, a reliable check for dates beyond the valid range is
    **     impossible.  To guard against gross errors, a year five or more
    **     after the release year of the present function (see parameter
    **     IYV) is considered dubious.  In this case a warning status is
    **     returned but the result is computed in the normal way.
    **
    **     For both too-early and too-late years, the warning status is
    **     j=+1.  This is distinct from the error status j=-1, which
    **     signifies a year so early that JD could not be computed.
    **
    **  2) If the specified date is for a day which ends with a leap second,
    **     the UTC-TAI value returned is for the period leading up to the
    **     leap second.  If the date is for a day which begins as a leap
    **     second ends, the UTC-TAI returned is for the period following the
    **     leap second.
    **
    **  3) The day number must be in the normal calendar range, for example
    **     1 through 30 for April.  The "almanac" convention of allowing
    **     such dates as January 0 and December 32 is not supported in this
    **     function, in order to avoid confusion near leap seconds.
    **
    **  4) The fraction of day is used only for dates before the
    **     introduction of leap seconds, the first of which occurred at the
    **     end of 1971.  It is tested for validity (zero to less than 1 is
    **     the valid range) even if not used;  if invalid, zero is used and
    **     status j=-4 is returned.  For many applications, setting fd to
    **     zero is acceptable;  the resulting error is always less than 3 ms
    **     (and occurs only pre-1972).
    **
    **  5) The status value returned in the case where there are multiple
    **     errors refers to the first error detected.  For example, if the
    **     month and day are 13 and 32 respectively, j=-2 (bad month)
    **     will be returned.
    **
    **  6) In cases where a valid result is not available, zero is returned.

    **                       1 = dubious year (Note 1)
    **                       0 = OK
    **                      -1 = bad year
    **                      -2 = bad month
    **                      -3 = bad day (Note 3)
    **                      -4 = bad fraction (Note 4)

    """
    cdef int i
    cdef int n = iy.shape[0]
    assert (iy.shape[0] == im.shape[0] == id.shape[0] == fd.shape[0])

    cdef np.ndarray[double, ndim=1] out = np.empty(n, dtype=np.double)

    warns = {1: DUBIOUS}
    errs = {-1: 'bad year',
             -2: 'bad month (must be 1 to 12)',
             -3: 'bad day (must be within normal calendar date for a month)',
             -4: 'bad fraction of day'}
    for i in range(n):
        ret = eraDat(iy[i], im[i], id[i], fd[i],
                     &out[i])
        check_return(ret, 'eraDat', warns, errs)

    return out


@cython.wraparound(False)
@cython.boundscheck(False)
def jd_dtf(scale, ndp,
              np.ndarray[double, ndim=1] d1,
              np.ndarray[double, ndim=1] d2):
    """
    int eraD2dtf(const char *scale, int ndp, double d1, double d2,
             int *iy, int *im, int *id, int ihmsf[4])

    **  Given:
    **     scale     char[]  time scale ID (Note 1)
    **     ndp       int     resolution (Note 2)
    **     d1,d2     double  time as a 2-part Julian Date (Notes 3,4)
    **
    **  Returned:
    **     iy,im,id  int     year, month, day in Gregorian calendar (Note 5)
    **     ihmsf     int[4]  hours, minutes, seconds, fraction (Note 1)
    **
    **  Returned (function value):
    **               int     status: +1 = dubious year (Note 5)
    **                                0 = OK
    **                               -1 = unacceptable date (Note 6)
    **
    **  Notes:
    **
    **  1) scale identifies the time scale.  Only the value "UTC" (in upper
    **     case) is significant, and enables handling of leap seconds (see
    **     Note 4).
    **
    **  2) ndp is the number of decimal places in the seconds field, and can
    **     have negative as well as positive values, such as:
    **
    **     ndp         resolution
    **     -4            1 00 00
    **     -3            0 10 00
    **     -2            0 01 00
    **     -1            0 00 10
    **      0            0 00 01
    **      1            0 00 00.1
    **      2            0 00 00.01
    **      3            0 00 00.001
    **
    **     The limits are platform dependent, but a safe range is -5 to +9.
    **
    **  3) d1+d2 is Julian Date, apportioned in any convenient way between
    **     the two arguments, for example where d1 is the Julian Day Number
    **     and d2 is the fraction of a day.  In the case of UTC, where the
    **     use of JD is problematical, special conventions apply:  see the
    **     next note.
    **
    **  4) JD cannot unambiguously represent UTC during a leap second unless
    **     special measures are taken.  The ERFA internal convention is that
    **     the quasi-JD day represents UTC days whether the length is 86399,
    **     86400 or 86401 SI seconds.
    **
    **  5) The warning status "dubious year" flags UTCs that predate the
    **     introduction of the time scale and that are too far in the future
    **     to be trusted.  See eraDat for further details.
    **
    **  6) For calendar conventions and limitations, see eraCal2jd.
    """
    cdef int i
    cdef int n = d1.shape[0]

    cdef np.ndarray[int, ndim=1] iy = np.empty(n, dtype=np.intc)
    cdef np.ndarray[int, ndim=1] im = np.empty(n, dtype=np.intc)
    cdef np.ndarray[int, ndim=1] id = np.empty(n, dtype=np.intc)
    cdef np.ndarray[int, ndim=2] ihmsf = np.empty((n, 4), dtype=np.intc)

    warns = {1: DUBIOUS}
    errs = {-1: 'unacceptable date'}

    for i in range(n):
        ret = eraD2dtf(scale, ndp, d1[i], d2[i],
                     &iy[i], &im[i], &id[i], &ihmsf[i, 0])
        check_return(ret, 'eraD2dtf', warns, errs)

    return iy, im, id, ihmsf

@cython.wraparound(False)
@cython.boundscheck(False)
def dtf_jd(scale,
              np.ndarray[int, ndim=1] iy,
              np.ndarray[int, ndim=1] im,
              np.ndarray[int, ndim=1] id,
              np.ndarray[int, ndim=1] ihr,
              np.ndarray[int, ndim=1] imn,
              np.ndarray[double, ndim=1] sec):
    """
    int eraDtf2d(char *scale, int iy, int im, int id, int ihr, int imn, double sec, double *d1, double *d2)

    **  Given:
    **     scale     char[]  time scale ID (Note 1)
    **     iy,im,id  int     year, month, day in Gregorian calendar (Note 2)
    **     ihr,imn   int     hour, minute
    **     sec       double  seconds
    **
    **  Returned:
    **     d1,d2     double  2-part Julian Date (Notes 3,4)
    **
    **  Returned (function value):
    **               int     status: +3 = both of next two
    **                               +2 = time is after end of day (Note 5)
    **                               +1 = dubious year (Note 6)
    **                                0 = OK
    **                               -1 = bad year
    **                               -2 = bad month
    **                               -3 = bad day
    **                               -4 = bad hour
    **                               -5 = bad minute
    **                               -6 = bad second (<0)
    **
    **  Notes:
    **
    **  1) scale identifies the time scale.  Only the value "UTC" (in upper
    **     case) is significant, and enables handling of leap seconds (see
    **     Note 4).
    **
    **  2) For calendar conventions and limitations, see eraCal2jd.
    **
    **  3) The sum of the results, d1+d2, is Julian Date, where normally d1
    **     is the Julian Day Number and d2 is the fraction of a day.  In the
    **     case of UTC, where the use of JD is problematical, special
    **     conventions apply:  see the next note.
    **
    **  4) JD cannot unambiguously represent UTC during a leap second unless
    **     special measures are taken.  The ERFA internal convention is that
    **     the quasi-JD day represents UTC days whether the length is 86399,
    **     86400 or 86401 SI seconds.
    **
    **  5) The warning status "time is after end of day" usually means that
    **     the sec argument is greater than 60.0.  However, in a day ending
    **     in a leap second the limit changes to 61.0 (or 59.0 in the case
    **     of a negative leap second).
    **
    **  6) The warning status "dubious year" flags UTCs that predate the
    **     introduction of the time scale and that are too far in the future
    **     to be trusted.  See eraDat for further details.
    **
    **  7) Only in the case of continuous and regular time scales (TAI, TT,
    **     TCG, TCB and TDB) is the result d1+d2 a Julian Date, strictly
    **     speaking.  In the other cases (UT1 and UTC) the result must be
    **     used with circumspection;  in particular the difference between
    **     two such results cannot be interpreted as a precise time
    **     interval.

    """
    cdef int i
    cdef int n = iy.shape[0]
    assert (iy.shape[0] == im.shape[0] == id.shape[0] ==
            ihr.shape[0] == imn.shape[0] == sec.shape[0])

    cdef np.ndarray[double, ndim=1] out1 = np.empty(n, dtype=np.double)
    cdef np.ndarray[double, ndim=1] out2 = np.empty(n, dtype=np.double)

    warns = {3: 'time is after end of day and ' + DUBIOUS,
             2: 'time is after end of day',
             1: DUBIOUS}
    errs = {-1: 'bad year',
            -2: 'bad month',
            -3: 'bad day',
            -4: 'bad hour',
            -5: 'bad minute',
            -6: 'bad second (< 0)'}

    for i in range(n):
        ret = eraDtf2d(scale, iy[i], im[i], id[i], ihr[i], imn[i], sec[i],
                       &out1[i], &out2[i])
        check_return(ret, 'eraDtf2d', warns, errs)

    return out1, out2


@cython.wraparound(False)
@cython.boundscheck(False)
def tai_tt( 
    np.ndarray[double, ndim=1] in1,
    np.ndarray[double, ndim=1] in2):
    """
    int eraTaitt(double tai1, double tai2, double *tt1, double *tt2)

    **  Given:
    **     tai1,tai2  double    TAI as a 2-part Julian Date
    **
    **  Returned:
    **     tt1,tt2    double    TT as a 2-part Julian Date
    **
    **  Returned (function value):
    **                int       status:  0 = OK
    **
    **  Note:
    **
    **     tai1+tai2 is Julian Date, apportioned in any convenient way
    **     between the two arguments, for example where tai1 is the Julian
    **     Day Number and tai2 is the fraction of a day.  The returned
    **     tt1,tt2 follow suit.
    """
    assert in1.shape[0] == in2.shape[0]
    cdef unsigned n = in1.shape[0]
    cdef unsigned int i
    cdef np.ndarray[double, ndim=1] out1 = np.empty(n, dtype=np.double)
    cdef np.ndarray[double, ndim=1] out2 = np.empty(n, dtype=np.double)

    for i in range(n):
        ret = eraTaitt(in1[i], in2[i], &out1[i], &out2[i])
        check_return(ret, 'eraTaitt')

    return out1, out2


@cython.wraparound(False)
@cython.boundscheck(False)
def tcb_tdb( 
    np.ndarray[double, ndim=1] in1,
    np.ndarray[double, ndim=1] in2):
    """
    int eraTcbtdb(double tcb1, double tcb2, double *tdb1, double *tdb2)

    **  Given:
    **     tcb1,tcb2  double    TCB as a 2-part Julian Date
    **
    **  Returned:
    **     tdb1,tdb2  double    TDB as a 2-part Julian Date
    **
    **  Returned (function value):
    **                int       status:  0 = OK
    **
    **  Notes:
    **
    **  1) tcb1+tcb2 is Julian Date, apportioned in any convenient way
    **     between the two arguments, for example where tcb1 is the Julian
    **     Day Number and tcb2 is the fraction of a day.  The returned
    **     tdb1,tdb2 follow suit.
    **
    **  2) The 2006 IAU General Assembly introduced a conventional linear
    **     transformation between TDB and TCB.  This transformation
    **     compensates for the drift between TCB and terrestrial time TT,
    **     and keeps TDB approximately centered on TT.  Because the
    **     relationship between TT and TCB depends on the adopted solar
    **     system ephemeris, the degree of alignment between TDB and TT over
    **     long intervals will vary according to which ephemeris is used.
    **     Former definitions of TDB attempted to avoid this problem by
    **     stipulating that TDB and TT should differ only by periodic
    **     effects.  This is a good description of the nature of the
    **     relationship but eluded precise mathematical formulation.  The
    **     conventional linear relationship adopted in 2006 sidestepped
    **     these difficulties whilst delivering a TDB that in practice was
    **     consistent with values before that date.
    **
    **  3) TDB is essentially the same as Teph, the time argument for the
    **     JPL solar system ephemerides.
    """

    assert in1.shape[0] == in2.shape[0]
    cdef unsigned n = in1.shape[0]
    cdef unsigned int i
    cdef np.ndarray[double, ndim=1] out1 = np.empty(n, dtype=np.double)
    cdef np.ndarray[double, ndim=1] out2 = np.empty(n, dtype=np.double)

    for i in range(n):
        ret = eraTcbtdb(in1[i], in2[i], &out1[i], &out2[i])
        check_return(ret, 'eraTcbtdb')

    return out1, out2


@cython.wraparound(False)
@cython.boundscheck(False)
def tcg_tt( 
    np.ndarray[double, ndim=1] in1,
    np.ndarray[double, ndim=1] in2):
    """
   int eraTcgtt(double tcg1, double tcg2, double *tt1, double *tt2)

   **  Given:
   **     tcg1,tcg2  double    TCG as a 2-part Julian Date
   **
   **  Returned:
   **     tt1,tt2    double    TT as a 2-part Julian Date
   **
   **  Returned (function value):
   **                int       status:  0 = OK
    """

    assert in1.shape[0] == in2.shape[0]
    cdef unsigned n = in1.shape[0]
    cdef unsigned int i
    cdef np.ndarray[double, ndim=1] out1 = np.empty(n, dtype=np.double)
    cdef np.ndarray[double, ndim=1] out2 = np.empty(n, dtype=np.double)

    for i in range(n):
        ret = eraTcgtt(in1[i], in2[i], &out1[i], &out2[i])
        check_return(ret, 'eraTcgtt')

    return out1, out2


@cython.wraparound(False)
@cython.boundscheck(False)
def tdb_tcb( 
    np.ndarray[double, ndim=1] in1,
    np.ndarray[double, ndim=1] in2):
    """
    int eraTdbtcb(double tdb1, double tdb2, double *tcb1, double *tcb2)

    **  Given:
    **     tdb1,tdb2  double    TDB as a 2-part Julian Date
    **
    **  Returned:
    **     tcb1,tcb2  double    TCB as a 2-part Julian Date
    **
    **  Returned (function value):
    **                int       status:  0 = OK
    """
    assert in1.shape[0] == in2.shape[0]
    cdef unsigned n = in1.shape[0]
    cdef unsigned int i
    cdef np.ndarray[double, ndim=1] out1 = np.empty(n, dtype=np.double)
    cdef np.ndarray[double, ndim=1] out2 = np.empty(n, dtype=np.double)

    for i in range(n):
        ret = eraTdbtcb(in1[i], in2[i], &out1[i], &out2[i])
        check_return(ret, 'eraTdbtcb')

    return out1, out2


@cython.wraparound(False)
@cython.boundscheck(False)
def tt_tai( 
    np.ndarray[double, ndim=1] in1,
    np.ndarray[double, ndim=1] in2):
    """
    int eraTttai(double tt1, double tt2, double *tai1, double *tai2)

    **  Given:
    **     tt1,tt2    double    TT as a 2-part Julian Date
    **
    **  Returned:
    **     tai1,tai2  double    TAI as a 2-part Julian Date
    **
    **  Returned (function value):
    **                int       status:  0 = OK
    """
    assert in1.shape[0] == in2.shape[0]
    cdef unsigned n = in1.shape[0]
    cdef unsigned int i
    cdef np.ndarray[double, ndim=1] out1 = np.empty(n, dtype=np.double)
    cdef np.ndarray[double, ndim=1] out2 = np.empty(n, dtype=np.double)

    for i in range(n):
        ret = eraTttai(in1[i], in2[i], &out1[i], &out2[i])
        check_return(ret, 'eraTttai')

    return out1, out2


@cython.wraparound(False)
@cython.boundscheck(False)
def tt_tcg( 
    np.ndarray[double, ndim=1] in1,
    np.ndarray[double, ndim=1] in2):
    """
    int eraTttcg(double tt1, double tt2, double *tcg1, double *tcg2)
    **  Given:
    **     tt1,tt2    double    TT as a 2-part Julian Date
    **
    **  Returned:
    **     tcg1,tcg2  double    TCG as a 2-part Julian Date
    **
    **  Returned (function value):
    **                int       status:  0 = OK
    """
    assert in1.shape[0] == in2.shape[0]
    cdef unsigned n = in1.shape[0]
    cdef unsigned int i
    cdef np.ndarray[double, ndim=1] out1 = np.empty(n, dtype=np.double)
    cdef np.ndarray[double, ndim=1] out2 = np.empty(n, dtype=np.double)

    for i in range(n):
        ret = eraTttcg(in1[i], in2[i], &out1[i], &out2[i])
        check_return(ret, 'eraTttcg')

    return out1, out2


@cython.wraparound(False)
@cython.boundscheck(False)
def utc_tai( 
    np.ndarray[double, ndim=1] in1,
    np.ndarray[double, ndim=1] in2):
    """
    int eraUtctai(double utc1, double utc2, double *tai1, double *tai2)

    **  Given:
    **     utc1,utc2  double   UTC as a 2-part quasi Julian Date (Notes 1-4)
    **
    **  Returned:
    **     tai1,tai2  double   TAI as a 2-part Julian Date (Note 5)
    **
    **  Returned (function value):
    **                int      status: +1 = dubious year (Note 3)
    **                                  0 = OK
    **                                 -1 = unacceptable date
    **
    **  Notes:
    **
    **  1) utc1+utc2 is quasi Julian Date (see Note 2), apportioned in any
    **     convenient way between the two arguments, for example where utc1
    **     is the Julian Day Number and utc2 is the fraction of a day.
    **
    **  2) JD cannot unambiguously represent UTC during a leap second unless
    **     special measures are taken.  The convention in the present
    **     function is that the JD day represents UTC days whether the
    **     length is 86399, 86400 or 86401 SI seconds.
    **
    **  3) The warning status "dubious year" flags UTCs that predate the
    **     introduction of the time scale and that are too far in the future
    **     to be trusted.  See eraDat  for further details.
    **
    **  4) The function eraDtf2d converts from calendar date and time of day
    **     into 2-part Julian Date, and in the case of UTC implements the
    **     leap-second-ambiguity convention described above.
    **
    **  5) The returned TAI1,TAI2 are such that their sum is the TAI Julian
    **     Date.
    **
    """
    assert in1.shape[0] == in2.shape[0]
    cdef unsigned n = in1.shape[0]
    cdef unsigned int i
    cdef np.ndarray[double, ndim=1] out1 = np.empty(n, dtype=np.double)
    cdef np.ndarray[double, ndim=1] out2 = np.empty(n, dtype=np.double)

    warns = {1: DUBIOUS}
    errs = {-1: 'unacceptable date'}

    for i in range(n):
        ret = eraUtctai(in1[i], in2[i], &out1[i], &out2[i])
        check_return(ret, 'eraUtctai', warns, errs)

    return out1, out2


@cython.wraparound(False)
@cython.boundscheck(False)
def tai_utc( 
    np.ndarray[double, ndim=1] in1,
    np.ndarray[double, ndim=1] in2):
    """
    int eraTaiutc(double tai1, double tai2, double *utc1, double *utc2)

    **  Given:
    **     tai1,tai2  double   TAI as a 2-part Julian Date (Note 1)
    **
    **  Returned:
    **     utc1,utc2  double   UTC as a 2-part quasi Julian Date (Notes 1-3)
    **
    **  Returned (function value):
    **                int      status: +1 = dubious year (Note 4)
    **                                  0 = OK
    **                                 -1 = unacceptable date
    **
    **  Notes:
    **
    **  1) tai1+tai2 is Julian Date, apportioned in any convenient way
    **     between the two arguments, for example where tai1 is the Julian
    **     Day Number and tai2 is the fraction of a day.  The returned utc1
    **     and utc2 form an analogous pair, except that a special convention
    **     is used, to deal with the problem of leap seconds - see the next
    **     note.
    **
    **  2) JD cannot unambiguously represent UTC during a leap second unless
    **     special measures are taken.  The convention in the present
    **     function is that the JD day represents UTC days whether the
    **     length is 86399, 86400 or 86401 SI seconds.
    **
    **  3) The function eraD2dtf can be used to transform the UTC quasi-JD
    **     into calendar date and clock time, including UTC leap second
    **     handling.
    **
    **  4) The warning status "dubious year" flags UTCs that predate the
    **     introduction of the time scale and that are too far in the future
    **     to be trusted.  See eraDat for further details.
        """
    assert in1.shape[0] == in2.shape[0]
    cdef unsigned n = in1.shape[0]
    cdef unsigned int i
    cdef np.ndarray[double, ndim=1] out1 = np.empty(n, dtype=np.double)
    cdef np.ndarray[double, ndim=1] out2 = np.empty(n, dtype=np.double)

    warns = {1: DUBIOUS}
    errs = {-1: 'unacceptable date'}

    for i in range(n):
        ret = eraTaiutc(in1[i], in2[i], &out1[i], &out2[i])
        check_return(ret, 'eraTaiutc', warns, errs)

    return out1, out2


@cython.wraparound(False)
@cython.boundscheck(False)
def tai_ut1( 
    np.ndarray[double, ndim=1] in1,
    np.ndarray[double, ndim=1] in2,
    np.ndarray[double, ndim=1] dt):
    """
    int eraTaiut1(double tai1, double tai2, double dta, double *ut11, double *ut12)

    **  Given:
    **     tai1,tai2  double    TAI as a 2-part Julian Date
    **     dta        double    UT1-TAI in seconds
    **
    **  Returned:
    **     ut11,ut12  double    UT1 as a 2-part Julian Date
    **
    **  Returned (function value):
    **                int       status:  0 = OK
    **
    **  Notes:
    **
    **  1) tai1+tai2 is Julian Date, apportioned in any convenient way
    **     between the two arguments, for example where tai1 is the Julian
    **     Day Number and tai2 is the fraction of a day.  The returned
    **     UT11,UT12 follow suit.
    **
    **  2) The argument dta, i.e. UT1-TAI, is an observed quantity, and is
    **     available from IERS tabulations.
    """
    assert in1.shape[0] == in2.shape[0] == dt.shape[0]
    cdef unsigned n = in1.shape[0]
    cdef unsigned int i
    cdef np.ndarray[double, ndim=1] out1 = np.empty(n, dtype=np.double)
    cdef np.ndarray[double, ndim=1] out2 = np.empty(n, dtype=np.double)

    for i in range(n):
        ret = eraTaiut1(in1[i], in2[i], dt[i], &out1[i], &out2[i])
        check_return(ret, 'eraTaiut1')

    return out1, out2


@cython.wraparound(False)
@cython.boundscheck(False)
def ut1_tai( 
    np.ndarray[double, ndim=1] in1,
    np.ndarray[double, ndim=1] in2,
    np.ndarray[double, ndim=1] dt):
    """
    int eraUt1tai(double ut11, double ut12, double dta, double *tai1, double *tai2)

    **  Given:
    **     ut11,ut12  double    UT1 as a 2-part Julian Date
    **     dta        double    UT1-TAI in seconds
    **
    **  Returned:
    **     tai1,tai2  double    TAI as a 2-part Julian Date
    **
    **  Returned (function value):
    **                int       status:  0 = OK
    **
    **  Notes:
    **
    **  1) ut11+ut12 is Julian Date, apportioned in any convenient way
    **     between the two arguments, for example where ut11 is the Julian
    **     Day Number and ut12 is the fraction of a day.  The returned
    **     tai1,tai2 follow suit.
    **
    **  2) The argument dta, i.e. UT1-TAI, is an observed quantity, and is
    **     available from IERS tabulations.
    """
    assert in1.shape[0] == in2.shape[0] == dt.shape[0]
    cdef unsigned n = in1.shape[0]
    cdef unsigned int i
    cdef np.ndarray[double, ndim=1] out1 = np.empty(n, dtype=np.double)
    cdef np.ndarray[double, ndim=1] out2 = np.empty(n, dtype=np.double)

    for i in range(n):
        ret = eraUt1tai(in1[i], in2[i], dt[i], &out1[i], &out2[i])
        check_return(ret, 'eraUt1tai')

    return out1, out2


@cython.wraparound(False)
@cython.boundscheck(False)
def tt_ut1( 
    np.ndarray[double, ndim=1] in1,
    np.ndarray[double, ndim=1] in2,
    np.ndarray[double, ndim=1] dt):
    """
    int eraTtut1(double tt1, double tt2, double dt, double *ut11, double *ut12)

    **  Given:
    **     tt1,tt2    double    TT as a 2-part Julian Date
    **     dt         double    TT-UT1 in seconds
    **
    **  Returned:
    **     ut11,ut12  double    UT1 as a 2-part Julian Date
    **
    **  Returned (function value):
    **                int       status:  0 = OK
    **
    **  Notes:
    **
    **  1) tt1+tt2 is Julian Date, apportioned in any convenient way between
    **     the two arguments, for example where tt1 is the Julian Day Number
    **     and tt2 is the fraction of a day.  The returned ut11,ut12 follow
    **     suit.
    **
    **  2) The argument dt is classical Delta T.
    """
    assert in1.shape[0] == in2.shape[0] == dt.shape[0]
    cdef unsigned n = in1.shape[0]
    cdef unsigned int i
    cdef np.ndarray[double, ndim=1] out1 = np.empty(n, dtype=np.double)
    cdef np.ndarray[double, ndim=1] out2 = np.empty(n, dtype=np.double)

    for i in range(n):
        ret = eraTtut1(in1[i], in2[i], dt[i], &out1[i], &out2[i])
        check_return(ret, 'eraTtut1')

    return out1, out2


@cython.wraparound(False)
@cython.boundscheck(False)
def ut1_tt( 
    np.ndarray[double, ndim=1] in1,
    np.ndarray[double, ndim=1] in2,
    np.ndarray[double, ndim=1] dt):
    """
    int eraUt1tt(double ut11, double ut12, double dt, double *tt1, double *tt2)

    **  Given:
    **     ut11,ut12  double    UT1 as a 2-part Julian Date
    **     dt         double    TT-UT1 in seconds
    **
    **  Returned:
    **     tt1,tt2    double    TT as a 2-part Julian Date
    **
    **  Returned (function value):
    **                int       status:  0 = OK
    **
    **  Notes:
    **
    **  1) ut11+ut12 is Julian Date, apportioned in any convenient way
    **     between the two arguments, for example where ut11 is the Julian
    **     Day Number and ut12 is the fraction of a day.  The returned
    **     tt1,tt2 follow suit.
    **
    **  2) The argument dt is classical Delta T.
    """
    assert in1.shape[0] == in2.shape[0] == dt.shape[0]
    cdef unsigned n = in1.shape[0]
    cdef unsigned int i
    cdef np.ndarray[double, ndim=1] out1 = np.empty(n, dtype=np.double)
    cdef np.ndarray[double, ndim=1] out2 = np.empty(n, dtype=np.double)

    for i in range(n):
        ret = eraUt1tt(in1[i], in2[i], dt[i], &out1[i], &out2[i])
        check_return(ret, 'eraUt1tt')

    return out1, out2


@cython.wraparound(False)
@cython.boundscheck(False)
def tdb_tt( 
    np.ndarray[double, ndim=1] in1,
    np.ndarray[double, ndim=1] in2,
    np.ndarray[double, ndim=1] dt):
    """
    int eraTdbtt(double tdb1, double tdb2, double dtr, double *tt1, double *tt2)

    **  Given:
    **     tdb1,tdb2  double    TDB as a 2-part Julian Date
    **     dtr        double    TDB-TT in seconds
    **
    **  Returned:
    **     tt1,tt2    double    TT as a 2-part Julian Date
    **
    **  Returned (function value):
    **                int       status:  0 = OK
    **
    **  Notes:
    **
    **  1) tdb1+tdb2 is Julian Date, apportioned in any convenient way
    **     between the two arguments, for example where tdb1 is the Julian
    **     Day Number and tdb2 is the fraction of a day.  The returned
    **     tt1,tt2 follow suit.
    **
    **  2) The argument dtr represents the quasi-periodic component of the
    **     GR transformation between TT and TCB.  It is dependent upon the
    **     adopted solar-system ephemeris, and can be obtained by numerical
    **     integration, by interrogating a precomputed time ephemeris or by
    **     evaluating a model such as that implemented in the ERFA function
    **     eraDtdb.   The quantity is dominated by an annual term of 1.7 ms
    **     amplitude.
    **
    **  3) TDB is essentially the same as Teph, the time argument for the
    **     JPL solar system ephemerides.
    """
    assert in1.shape[0] == in2.shape[0] == dt.shape[0]
    cdef unsigned n = in1.shape[0]
    cdef unsigned int i
    cdef np.ndarray[double, ndim=1] out1 = np.empty(n, dtype=np.double)
    cdef np.ndarray[double, ndim=1] out2 = np.empty(n, dtype=np.double)

    for i in range(n):
        ret = eraTdbtt(in1[i], in2[i], dt[i], &out1[i], &out2[i])
        check_return(ret, 'eraTdbtt')

    return out1, out2


@cython.wraparound(False)
@cython.boundscheck(False)
def tt_tdb( 
    np.ndarray[double, ndim=1] in1,
    np.ndarray[double, ndim=1] in2,
    np.ndarray[double, ndim=1] dt):
    """
    int eraTttdb(double tt1, double tt2, double dtr, double *tdb1, double *tdb2)

    **  Given:
    **     tt1,tt2    double    TT as a 2-part Julian Date
    **     dtr        double    TDB-TT in seconds
    **
    **  Returned:
    **     tdb1,tdb2  double    TDB as a 2-part Julian Date
    **
    **  Returned (function value):
    **                int       status:  0 = OK
    **
    **  Notes:
    **
    **  1) tt1+tt2 is Julian Date, apportioned in any convenient way between
    **     the two arguments, for example where tt1 is the Julian Day Number
    **     and tt2 is the fraction of a day.  The returned tdb1,tdb2 follow
    **     suit.
    **
    **  2) The argument dtr represents the quasi-periodic component of the
    **     GR transformation between TT and TCB.  It is dependent upon the
    **     adopted solar-system ephemeris, and can be obtained by numerical
    **     integration, by interrogating a precomputed time ephemeris or by
    **     evaluating a model such as that implemented in the ERFA function
    **     eraDtdb.   The quantity is dominated by an annual term of 1.7 ms
    **     amplitude.
    **
    **  3) TDB is essentially the same as Teph, the time argument for the JPL
    **     solar system ephemerides.
    """
    assert in1.shape[0] == in2.shape[0] == dt.shape[0]
    cdef unsigned n = in1.shape[0]
    cdef unsigned int i
    cdef np.ndarray[double, ndim=1] out1 = np.empty(n, dtype=np.double)
    cdef np.ndarray[double, ndim=1] out2 = np.empty(n, dtype=np.double)

    for i in range(n):
        ret = eraTttdb(in1[i], in2[i], dt[i], &out1[i], &out2[i])
        check_return(ret, 'eraTttdb')

    return out1, out2


@cython.wraparound(False)
@cython.boundscheck(False)
def ut1_utc( 
    np.ndarray[double, ndim=1] in1,
    np.ndarray[double, ndim=1] in2,
    np.ndarray[double, ndim=1] dt):
    """
    int eraUt1utc(double ut11, double ut12, double dut1, double *utc1, double *utc2)

    **  Given:
    **     ut11,ut12  double   UT1 as a 2-part Julian Date (Note 1)
    **     dut1       double   Delta UT1: UT1-UTC in seconds (Note 2)
    **
    **  Returned:
    **     utc1,utc2  double   UTC as a 2-part quasi Julian Date (Notes 3,4)
    **
    **  Returned (function value):
    **                int      status: +1 = dubious year (Note 5)
    **                                  0 = OK
    **                                 -1 = unacceptable date
    **
    **  Notes:
    **
    **  1) ut11+ut12 is Julian Date, apportioned in any convenient way
    **     between the two arguments, for example where ut11 is the Julian
    **     Day Number and ut12 is the fraction of a day.  The returned utc1
    **     and utc2 form an analogous pair, except that a special convention
    **     is used, to deal with the problem of leap seconds - see Note 3.
    **
    **  2) Delta UT1 can be obtained from tabulations provided by the
    **     International Earth Rotation and Reference Systems Service.  The
    **     value changes abruptly by 1s at a leap second;  however, close to
    **     a leap second the algorithm used here is tolerant of the "wrong"
    **     choice of value being made.
    **
    **  3) JD cannot unambiguously represent UTC during a leap second unless
    **     special measures are taken.  The convention in the present
    **     function is that the returned quasi JD day UTC1+UTC2 represents
    **     UTC days whether the length is 86399, 86400 or 86401 SI seconds.
    **
    **  4) The function eraD2dtf can be used to transform the UTC quasi-JD
    **     into calendar date and clock time, including UTC leap second
    **     handling.
    **
    **  5) The warning status "dubious year" flags UTCs that predate the
    **     introduction of the time scale and that are too far in the future
    **     to be trusted.  See eraDat for further details.
    """
    assert in1.shape[0] == in2.shape[0] == dt.shape[0]
    cdef unsigned n = in1.shape[0]
    cdef unsigned int i
    cdef np.ndarray[double, ndim=1] out1 = np.empty(n, dtype=np.double)
    cdef np.ndarray[double, ndim=1] out2 = np.empty(n, dtype=np.double)

    warns = {1: DUBIOUS}
    errs = {-1: 'unacceptable date'}

    for i in range(n):
        ret = eraUt1utc(in1[i], in2[i], dt[i], &out1[i], &out2[i])
        check_return(ret, 'eraUt1utc', warns, errs)

    return out1, out2


@cython.wraparound(False)
@cython.boundscheck(False)
def utc_ut1( 
    np.ndarray[double, ndim=1] in1,
    np.ndarray[double, ndim=1] in2,
    np.ndarray[double, ndim=1] dt):
    """
    int eraUtcut1(double utc1, double utc2, double dut1, double *ut11, double *ut12)

    **  Given:
    **     utc1,utc2  double   UTC as a 2-part quasi Julian Date (Notes 1-4)
    **     dut1       double   Delta UT1 = UT1-UTC in seconds (Note 5)
    **
    **  Returned:
    **     ut11,ut12  double   UT1 as a 2-part Julian Date (Note 6)
    **
    **  Returned (function value):
    **                int      status: +1 = dubious year (Note 7)
    **                                  0 = OK
    **                                 -1 = unacceptable date
    **
    **  Notes:
    **
    **  1) utc1+utc2 is quasi Julian Date (see Note 2), apportioned in any
    **     convenient way between the two arguments, for example where utc1
    **     is the Julian Day Number and utc2 is the fraction of a day.
    **
    **  2) JD cannot unambiguously represent UTC during a leap second unless
    **     special measures are taken.  The convention in the present
    **     function is that the JD day represents UTC days whether the
    **     length is 86399, 86400 or 86401 SI seconds.
    **
    **  3) The warning status "dubious year" flags UTCs that predate the
    **     introduction of the time scale and that are too far in the future
    **     to be trusted.  See eraDat  for further details.
    **
    **  4) The function eraDtf2d  converts from calendar date and time of
    **     day into 2-part Julian Date, and in the case of UTC implements
    **     the leap-second-ambiguity convention described above.
    **
    **  5) Delta UT1 can be obtained from tabulations provided by the
    **     International Earth Rotation and Reference Systems Service.  It
    **     It is the caller's responsibility to supply a DUT argument
    **     containing the UT1-UTC value that matches the given UTC.
    **
    **  6) The returned ut11,ut12 are such that their sum is the UT1 Julian
    **     Date.
    **
    **  7) The warning status "dubious year" flags UTCs that predate the
    **     introduction of the time scale and that are too far in the future
    **     to be trusted.  See eraDat for further details.
    """
    assert in1.shape[0] == in2.shape[0] == dt.shape[0]
    cdef unsigned n = in1.shape[0]
    cdef unsigned int i
    cdef np.ndarray[double, ndim=1] out1 = np.empty(n, dtype=np.double)
    cdef np.ndarray[double, ndim=1] out2 = np.empty(n, dtype=np.double)

    warns = {1: DUBIOUS}
    errs = {-1: 'unacceptable date'}

    for i in range(n):
        ret = eraUtcut1(in1[i], in2[i], dt[i], &out1[i], &out2[i])
        check_return(ret, 'eraUtcut1', warns, errs)

    return out1, out2


@cython.wraparound(False)
@cython.boundscheck(False)
def d_tdb_tt(np.ndarray[double, ndim=1] in1,
             np.ndarray[double, ndim=1] in2,
             np.ndarray[double, ndim=1] ut,
             elong, u, v):
    """
    compute DTR = TDB-TT
    double eraDtdb(double date1, double date2, double ut,
                   double elong, double u, double v)

    **  Given:
    **     date1,date2   double  date, TDB (Notes 1-3)
    **     ut            double  universal time (UT1, fraction of one day)
    **     elong         double  longitude (east positive, radians)
    **     u             double  distance from Earth spin axis (km)
    **     v             double  distance north of equatorial plane (km)
    **
    **  Returned (function value):
    **                   double  TDB-TT (seconds)
    **
    **  Notes:
    **
    **  1) The date date1+date2 is a Julian Date, apportioned in any
    **     convenient way between the two arguments.  For example,
    **     JD(TT)=2450123.7 could be expressed in any of these ways,
    **     among others:
    **
    **            date1          date2
    **
    **         2450123.7           0.0       (JD method)
    **         2451545.0       -1421.3       (J2000 method)
    **         2400000.5       50123.2       (MJD method)
    **         2450123.5           0.2       (date & time method)
    **
    **     The JD method is the most natural and convenient to use in
    **     cases where the loss of several decimal digits of resolution
    **     is acceptable.  The J2000 method is best matched to the way
    **     the argument is handled internally and will deliver the
    **     optimum resolution.  The MJD method and the date & time methods
    **     are both good compromises between resolution and convenience.
    **
    **     Although the date is, formally, barycentric dynamical time (TDB),
    **     the terrestrial dynamical time (TT) can be used with no practical
    **     effect on the accuracy of the prediction.
    **
    **  2) TT can be regarded as a coordinate time that is realized as an
    **     offset of 32.184s from International Atomic Time, TAI.  TT is a
    **     specific linear transformation of geocentric coordinate time TCG,
    **     which is the time scale for the Geocentric Celestial Reference
    **     System, GCRS.
    **
    **  3) TDB is a coordinate time, and is a specific linear transformation
    **     of barycentric coordinate time TCB, which is the time scale for
    **     the Barycentric Celestial Reference System, BCRS.
    **
    **  4) The difference TCG-TCB depends on the masses and positions of the
    **     bodies of the solar system and the velocity of the Earth.  It is
    **     dominated by a rate difference, the residual being of a periodic
    **     character.  The latter, which is modeled by the present function,
    **     comprises a main (annual) sinusoidal term of amplitude
    **     approximately 0.00166 seconds, plus planetary terms up to about
    **     20 microseconds, and lunar and diurnal terms up to 2 microseconds.
    **     These effects come from the changing transverse Doppler effect
    **     and gravitational red-shift as the observer (on the Earth's
    **     surface) experiences variations in speed (with respect to the
    **     BCRS) and gravitational potential.
    **
    **  5) TDB can be regarded as the same as TCB but with a rate adjustment
    **     to keep it close to TT, which is convenient for many applications.
    **     The history of successive attempts to define TDB is set out in
    **     Resolution 3 adopted by the IAU General Assembly in 2006, which
    **     defines a fixed TDB(TCB) transformation that is consistent with
    **     contemporary solar-system ephemerides.  Future ephemerides will
    **     imply slightly changed transformations between TCG and TCB, which
    **     could introduce a linear drift between TDB and TT;  however, any
    **     such drift is unlikely to exceed 1 nanosecond per century.
    **
    **  6) The geocentric TDB-TT model used in the present function is that of
    **     Fairhead & Bretagnon (1990), in its full form.  It was originally
    **     supplied by Fairhead (private communications with P.T.Wallace,
    **     1990) as a Fortran subroutine.  The present C function contains an
    **     adaptation of the Fairhead code.  The numerical results are
    **     essentially unaffected by the changes, the differences with
    **     respect to the Fairhead & Bretagnon original being at the 1e-20 s
    **     level.
    **
    **     The topocentric part of the model is from Moyer (1981) and
    **     Murray (1983), with fundamental arguments adapted from
    **     Simon et al. 1994.  It is an approximation to the expression
    **     ( v / c ) . ( r / c ), where v is the barycentric velocity of
    **     the Earth, r is the geocentric position of the observer and
    **     c is the speed of light.
    **
    **     By supplying zeroes for u and v, the topocentric part of the
    **     model can be nullified, and the function will return the Fairhead
    **     & Bretagnon result alone.
    **
    **  7) During the interval 1950-2050, the absolute accuracy is better
    **     than +/- 3 nanoseconds relative to time ephemerides obtained by
    **     direct numerical integrations based on the JPL DE405 solar system
    **     ephemeris.
    **
    **  8) It must be stressed that the present function is merely a model,
    **     and that numerical integration of solar-system ephemerides is the
    **     definitive method for predicting the relationship between TCG and
    **     TCB and hence between TT and TDB.
    """
    assert in1.shape[0] == in2.shape[0] == ut.shape[0]
    cdef unsigned n = in1.shape[0]
    cdef unsigned int i
    cdef np.ndarray[double, ndim=1] out = np.empty(n, dtype=np.double)
    cdef double c_elong = elong
    cdef double c_u = u
    cdef double c_v = v

    for i in range(n):
        out[i] = eraDtdb(in1[i], in2[i], ut[i], c_elong, c_u, c_v)
    return out


def era_af2a(sign, ideg, iamin, asec):
    """
    int eraAf2a(char s, int ideg, int iamin, double asec, double *rad)

    **  Given:
    **     s         char    sign:  '-' = negative, otherwise positive
    **     ideg      int     degrees
    **     iamin     int     arcminutes
    **     asec      double  arcseconds
    **
    **  Returned:
    **     rad       double  angle in radians
    **
    **  Returned (function value):
    **               int     status:  0 = OK
    **                                1 = ideg outside range 0-359
    **                                2 = iamin outside range 0-59
    **                                3 = asec outside range 0-59.999...
    **
    **  Notes:
    **
    **  1)  The result is computed even if any of the range checks fail.
    **
    **  2)  Negative ideg, iamin and/or asec produce a warning status, but
    **      the absolute value is used in the conversion.
    **
    **  3)  If there are multiple errors, the status value reflects only the
    **      first, the smallest taking precedence.
    """
    cdef double rad
    s = ord(sign)

    warns = {1: 'ideg outside range 0-359',
             2: 'iamin outside range 0-59',
             3: 'asec outside range 0-59.999...'}

    ret = eraAf2a(s, ideg, iamin, asec, &rad)
    check_return(ret, 'eraAf2a', warns)

    return rad

def era_gd2gc(n, elong, phi, height):
    """
    Wrap
    int eraGd2gc(int n, double elong, double phi, double height, double xyz[3])

    **  Given:
    **     n       int        ellipsoid identifier (Note 1)
    **     elong   double     longitude (radians, east +ve)
    **     phi     double     latitude (geodetic, radians, Note 3)
    **     height  double     height above ellipsoid (geodetic, Notes 2,3)
    **
    **  Returned:
    **     xyz     double[3]  geocentric vector (Note 2)
    **
    **  Returned (function value):
    **             int        status:  0 = OK
    **                                -1 = illegal identifier (Note 3)
    **                                -2 = illegal case (Note 3)
    **
    **  Notes:
    **
    **  1) The identifier n is a number that specifies the choice of
    **     reference ellipsoid.  The following are supported:
    **
    **        n    ellipsoid
    **
    **        1     WGS84
    **        2     GRS80
    **        3     WGS72
    **
    **     The n value has no significance outside the ERFA software.  For
    **     convenience, symbols WGS84 etc. are defined in erfam.h.
    **
    **  2) The height (height, given) and the geocentric vector (xyz,
    **     returned) are in meters.
    **
    **  3) No validation is performed on the arguments elong, phi and
    **     height.  An error status -1 means that the identifier n is
    **     illegal.  An error status -2 protects against cases that would
    **     lead to arithmetic exceptions.  In all error cases, xyz is set
    **     to zeros.
    **
    **  4) The inverse transformation is performed in the function eraGc2gd.
    """
    cdef np.ndarray[double, ndim=1] xyz = np.empty(3, dtype=np.double)

    errs = {-1: 'illegal identifier',
             -2: 'illegal case'}

    ret = eraGd2gc(n, elong, phi, height, &xyz[0])
    check_return(ret, 'eraGd2gc', errors=errs)

    return xyz


@cython.wraparound(False)
@cython.boundscheck(False)
def jd_julian_epoch(np.ndarray[double, ndim=1] jd1,
                    np.ndarray[double, ndim=1] jd2):
    """ Wrap double eraEpj(double dj1, double dj2)
    **  Julian Date to Julian Epoch.

    **  Given:
    **     dj1,dj2    double     Julian Date (see note)
    **
    **  Returned (function value):
    **                double     Julian Epoch
    **
    **  Note:
    **
    **     The Julian Date is supplied in two pieces, in the usual ERFA
    **     manner, which is designed to preserve time resolution.  The
    **     Julian Date is available as a single number by adding dj1 and
    **     dj2.  The maximum resolution is achieved if dj1 is 2451545D0
    **     (J2000.0).
    """
    assert jd1.shape[0] == jd2.shape[0]
    cdef unsigned n = jd1.shape[0]
    cdef unsigned int i
    cdef np.ndarray[double, ndim=1] epd = np.empty(n, dtype=np.double)

    for i in range(n):
        epd[i] = eraEpj(jd1[i], jd2[i])
    return epd


@cython.wraparound(False)
@cython.boundscheck(False)
def julian_epoch_jd(np.ndarray[double, ndim=1] epd):
    """ Wrap void eraEpj2jd(double epj, double *djm0, double *djm)
    **  Julian Epoch to Julian Date.
    **  Given:
    **     epj      double    Julian Epoch (e.g. 1996.8D0)
    **
    **  Returned:
    **     djm0     double    MJD zero-point: always 2400000.5
    **     djm      double    Modified Julian Date
    """
    cdef unsigned n = epd.shape[0]
    cdef unsigned int i
    cdef np.ndarray[double, ndim=1] jd1 = np.empty(n, dtype=np.double)
    cdef np.ndarray[double, ndim=1] jd2 = np.empty(n, dtype=np.double)

    for i in range(n):
        eraEpj2jd(epd[i], &jd1[i], &jd2[i])
    return jd1, jd2


@cython.wraparound(False)
@cython.boundscheck(False)
def jd_besselian_epoch(np.ndarray[double, ndim=1] jd1,
                       np.ndarray[double, ndim=1] jd2):
    """ Wrap double eraEpb(double dj1, double dj2)
    **  Julian Date to Besselian Epoch.

    **  Given:
    **     dj1,dj2    double     Julian Date (see note)
    **
    **  Returned (function value):
    **                double     Besselian Epoch.
    **
    **  Note:
    **
    **     The Julian Date is supplied in two pieces, in the usual ERFA
    **     manner, which is designed to preserve time resolution.  The
    **     Julian Date is available as a single number by adding dj1 and
    **     dj2.  The maximum resolution is achieved if dj1 is 2451545D0
    **     (J2000.0).
    """
    assert jd1.shape[0] == jd2.shape[0]
    cdef unsigned n = jd1.shape[0]
    cdef unsigned int i
    cdef np.ndarray[double, ndim=1] epd = np.empty(n, dtype=np.double)

    for i in range(n):
        epd[i] = eraEpb(jd1[i], jd2[i])
    return epd


@cython.wraparound(False)
@cython.boundscheck(False)
def besselian_epoch_jd(np.ndarray[double, ndim=1] epd):
    """ Wrap void eraEpb2jd(double epj, double *djm0, double *djm)
    **  Besselian Epoch to Julian Date.

    **  Given:
    **     epb      double    Besselian Epoch (e.g. 1957.3D0)
    **
    **  Returned:
    **     djm0     double    MJD zero-point: always 2400000.5
    **     djm      double    Modified Julian Date
    **
    **  Note:
    **
    **     The Julian Date is returned in two pieces, in the usual ERFA
    **     manner, which is designed to preserve time resolution.  The
    **     Julian Date is available as a single number by adding djm0 and
    **     djm.
    """
    cdef unsigned n = epd.shape[0]
    cdef unsigned int i
    cdef np.ndarray[double, ndim=1] jd1 = np.empty(n, dtype=np.double)
    cdef np.ndarray[double, ndim=1] jd2 = np.empty(n, dtype=np.double)

    for i in range(n):
        eraEpb2jd(epd[i], &jd1[i], &jd2[i])
    return jd1, jd2



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This module defines base classes for all models.
The base class of all models is `~astropy.modeling.Model`.
`~astropy.modeling.ParametricModel` is the base class for all fittable models. Parametric
models can be linear or nonlinear in a regression analysis sense.

All models provide a `__call__` method which performs the transformation in a
purely mathematical way, i.e. the models are unitless. In addition, when
possible the transformation is done using multiple parameter sets, `param_sets`.
The number of parameter sets is stored in an attribute `param_dim`.

Parametric models also store a flat list of all parameters as an instance of
`~astropy.modeling.parameters.Parameters`. When fitting, this list-like object is modified by a
subclass of `~astropy.modeling.fitting.Fitter`. When fitting nonlinear models, the values of the
parameters are used as initial guesses by the fitting class. Normally users
will not have to use the `~astropy.modeling.parameters` module directly.

Input Format For Model Evaluation and Fitting

Input coordinates are passed in separate arguments, for example 2D models
expect x and y coordinates to be passed separately as two scalars or aray-like
objects.
The evaluation depends on the input dimensions and the number of parameter
sets but in general normal broadcasting rules apply.
For example:

- A model with one parameter set works with input in any dimensionality

- A model with N parameter sets works with 2D arrays of shape (M, N).
  A parameter set is applied to each column.

- A model with N parameter sets works with multidimensional arrays if the
  shape of the input array is (N, M, P). A parameter set is applied to each plane.

In all these cases the output has the same shape as the input.

- A model with N parameter sets works with 1D input arrays. The shape
  of the output is (M, N)
"""

from __future__ import division

import abc
import functools
import copy
from itertools import izip
from textwrap import dedent

import numpy as np

from .parameters import Parameter, InputParameterError
from ..utils import indent, isiterable


__all__ = ['Model', 'ParametricModel', 'SummedCompositeModel',
           'SerialCompositeModel', 'LabeledInput', 'Parametric1DModel',
           'Parametric2DModel', 'ModelDefinitionError', 'format_input']


class ModelDefinitionError(Exception):
    """Used for incorrect models definitions"""


def format_input(func):
    """
    Wraps a model's ``__call__`` method so that the input arrays are converted
    into the appropriate shape given the model's parameter dimensions.

    Wraps the result to match the shape of the last input array.
    """

    @functools.wraps(func)
    def wrapped_call(self, *args):
        converted = []

        for arg in args:
            # Reset these flags; their value only matters for the last
            # argument
            transposed = False
            scalar = False

            arg = np.asarray(arg) + 0.
            if self.param_dim == 1:
                if arg.ndim == 0:
                    scalar = True
                converted.append(arg)
                continue

            if arg.ndim < 2:
                converted.append(np.array([arg]).T)
            elif arg.ndim == 2:
                assert arg.shape[-1] == self.param_dim, \
                    ("Cannot broadcast with shape "
                     "({0}, {1})".format(arg.shape[0], arg.shape[1]))
                converted.append(arg)
            elif arg.ndim > 2:
                assert arg.shape[0] == self.param_dim, \
                    ("Cannot broadcast with shape "
                     "({0}, {1}, {2})".format(arg.shape[0], arg.shape[1],
                                              arg.shape[2]))
                transposed = True
                converted.append(arg.T)

        result = func(self, *converted)

        if transposed:
            return result.T
        elif scalar:
            try:
                return result[0]
            except IndexError:
                return result

        return result

    return wrapped_call


class _ModelMeta(abc.ABCMeta):
    """
    Metaclass for Model.

    Currently just handles auto-generating the param_names list based on
    Parameter descriptors declared at the class-level of Model subclasses.
    """

    def __new__(mcls, name, bases, members):
        param_names = members.get('param_names', [])
        parameters = dict((value.name, value) for value in members.values()
                          if isinstance(value, Parameter))

        # If no parameters were defined get out early--this is especially
        # important for PolynomialModels which take a different approach to
        # parameters, since they can have a variable number of them
        if not parameters:
            return super(_ModelMeta, mcls).__new__(mcls, name, bases, members)

        # If param_names was declared explicitly we use only the parameters
        # listed manually in param_names, but still check that all listed
        # parameters were declared
        if param_names and isiterable(param_names):
            for param_name in param_names:
                if param_name not in parameters:
                    raise RuntimeError(
                        "Parameter {0!r} listed in {1}.param_names was not "
                        "declared in the class body.".format(param_name, name))
        else:
            param_names = [param.name for param in
                           sorted(parameters.values(),
                                  key=lambda p: p._order)]
            members['param_names'] = param_names

        return super(_ModelMeta, mcls).__new__(mcls, name, bases, members)


class Model(object):
    """
    Base class for all models.

    This is an abstract class and should not be instantiated directly.

    Notes
    -----
    Models which are not meant to be fit to data should subclass this class.

    This class sets the properties for all individual parameters and performs
    parameter validation.
    """

    __metaclass__ = _ModelMeta

    param_names = []
    n_inputs = 1
    n_outputs = 1
    fittable = False
    linear = True

    def __init__(self, param_dim=1):
        self._param_dim = param_dim

    @property
    def param_dim(self):
        """Number of parameter sets in a model."""

        return self._param_dim

    def __repr__(self):
        fmt = "{0}(".format(self.__class__.__name__)
        for name in self.param_names:
            fmt1 = """
            {0}={1},
            """.format(name, getattr(self, name))
            fmt += fmt1
        fmt += ")"

        return fmt

    def __str__(self):
        fmt = """
        Model: {0}
        Parameter sets: {1}
        Parameters: \n{2}
        """.format(
              self.__class__.__name__,
              self.param_dim,
              indent('\n'.join('{0}: {1}'.format(n, getattr(self, n))
                               for n in self.param_names),
                     width=19))

        return dedent(fmt[1:])

    @property
    def param_sets(self):
        """
        Return parameters as a pset.

        This is an array where each column represents one parameter set.
        """

        parameters = [getattr(self, attr) for attr in self.param_names]
        values = [par.value for par in parameters]
        shapes = [par.shape for par in parameters]
        n_dims = np.asarray([len(p.shape) for p in parameters])

        if (n_dims > 1).any():
            if () in shapes:
                psets = np.asarray(values, dtype=np.object)
            else:
                psets = np.asarray(values)
        else:
            psets = np.asarray(values)
            psets.shape = (len(self.param_names), self.param_dim)
        return psets

    def inverse(self):
        """Returns a callable object which performs the inverse transform."""

        raise NotImplementedError("An analytical inverse transform has not "
                                  "been implemented for this model.")

    def invert(self):
        """Invert coordinates iteratively if possible."""

        raise NotImplementedError("Subclasses should implement this")

    def add_model(self, newtr, mode):
        """
        Create a CompositeModel by chaining the current model with the new one
        using the specified mode.

        Parameters
        ----------
        newtr : an instance of a subclass of Model
        mode :  string
               'parallel', 'serial', 'p' or 's'
               a flag indicating whether to combine the models
               in series or in parallel

        Returns
        -------
        model : CompositeModel
            an instance of CompositeModel
        """

        if mode in ['parallel', 'p']:
            return SummedCompositeModel([self, newtr])
        elif mode in ['serial', 's']:
            return SerialCompositeModel([self, newtr])
        else:
            raise InputParameterError("Unrecognized mode {0}".format(mode))

    def copy(self):
        return copy.deepcopy(self)

    @abc.abstractmethod
    def __call__(self):
        raise NotImplementedError("Subclasses should implement this")


class ParametricModel(Model):
    """
    Base class for all fittable models.

    Notes
    -----
    All models which can be fit to data should subclass this class.

    Sets the ``parameters`` attributes.

    Parameters
    ----------
    param_dim : int
        number of parameter sets
    fittable : bool
        indicator if the model is fittable
    fixed : a dict
        a dictionary ``{parameter_name: boolean}`` of parameters to not be
        varied during fitting. True means the parameter is held fixed.
        Alternatively the `~astropy.modeling.parameters.Parameter.fixed`
        property of a parameter may be used.
    tied : dict
        a dictionary ``{parameter_name: callable}`` of parameters which are
        linked to some other parameter. The dictionary values are callables
        providing the linking relationship.
        Alternatively the `~astropy.modeling.parameters.Parameter.tied`
        property of a parameter may be used.
    bounds : dict
        a dictionary ``{parameter_name: boolean}`` of lower and upper bounds of
        parameters. Keys are parameter names. Values are a list of length 2
        giving the desired range for the parameter.  Alternatively the
        `~astropy.modeling.parameters.Parameter.min` and
        `~astropy.modeling.parameters.Parameter.max` properties of a parameter
        may be used.
    eqcons : list
        A list of functions of length n such that
        ``eqcons[j](x0, *args) == 0.0`` in a successfully optimized
        problem.
    ineqcons : list
        A list of functions of length n such that
        ``ieqcons[j](x0, *args) >= 0.0`` is a successfully optimized
        problem.

    Examples
    --------
    >>> from astropy.modeling import models
    >>> def tie_center(model):
    ...         mean = 50 * model.stddev
    ...         return mean
    >>> tied_parameters = {'mean': tie_center}

    Specify that ``'mean'`` is a tied parameter in one of two ways:

    >>> g1 = models.Gaussian1D(amplitude=10, mean=5, stddev=.3,
    ...                             tied=tied_parameters)

    or

    >>> g1 = models.Gaussian1D(amplitude=10, mean=5, stddev=.3)
    >>> g1.mean.tied
    False
    >>> g1.mean.tied = tie_center
    >>> g1.mean.tied
    <function tie_center at 0x...>

    Fixed parameters:

    >>> g1 = models.Gaussian1D(amplitude=10, mean=5, stddev=.3,
    ...                             fixed={'stddev': True})
    >>> g1.stddev.fixed
    True

    or

    >>> g1 = models.Gaussian1D(amplitude=10, mean=5, stddev=.3)
    >>> g1.stddev.fixed
    False
    >>> g1.stddev.fixed = True
    >>> g1.stddev.fixed
    True
    """

    linear = False
    deriv = None
    # Flag that indicates if the model derivatives are given in columns
    # or rows
    col_deriv = True
    fittable = True


    def __init__(self, **kwargs):
        # Pop any constraints off the keyword arguments
        bounds = kwargs.pop('bounds', None)
        fixed = kwargs.pop('fixed', None)
        tied = kwargs.pop('tied', None)
        eqcons = kwargs.pop('eqcons', None)
        ineqcons = kwargs.pop('ineqcons', None)

        # Pop off the param_dims
        param_dim = kwargs.pop('param_dim', None)

        # Remaining keyword args are either parameter values or invalid
        # Parameter values must be passed in as keyword arguments in order to
        # distinguish them
        params = kwargs

        # Determine the number of parameter sets: This will be based
        # on the size of any parameters whose values have been specified
        # or the default of 1 is used
        # This loop also checks that all the supplied parameter names are
        # valid
        param_names = set(self.param_names)

        max_param_dim = 1
        for name, value in params.items():
            if param_dim is None:
                # Determine the best param_dims, if not already specified,
                # based on the sizes of the input parameter values
                max_param_dim = max(max_param_dim, np.size(value))

            if name not in param_names:
                raise ValueError(
                    "Unrecognized parameter: {0}".format(name))

        if param_dim is None:
            param_dim = max_param_dim

        super(ParametricModel, self).__init__(param_dim=param_dim)

        # Initialize the constraints for each parameter
        if eqcons is None:
            self._eqcons = []
        else:
            self._eqcons = eqcons
        if ineqcons is None:
            self._ineqcons = []
        else:
            self._ineqcons = ineqcons

        # Set constraints
        self._constraints = {}

        if fixed:
            self._constraints['fixed'] = fixed
        if tied:
            self._constraints['tied'] = tied
        if bounds:
            self._constraints['bounds'] = bounds

        self._initialize_parameters(params)


    @property
    def fixed(self):
        """
        A dictionary mapping parameter names to their fixed constraint
        """

        return dict((name, getattr(self, name).fixed)
                    for name in self.param_names)

    @property
    def tied(self):
        """
        A dictionary mapping parameter names to their tied constraint
        """

        return dict((name, getattr(self, name).tied)
                    for name in self.param_names)

    @property
    def bounds(self):
        return dict((name, getattr(self, name).bounds)
                    for name in self.param_names)

    @property
    def eqcons(self):
        return self._eqcons

    @property
    def ineqcons(self):
        return self._ineqcons

    @property
    def parameters(self):
        """
        A flattened array of all parameter values in all parameter sets

        Fittable parameters maintain this list and fitters modify it.
        """

        return self._parameters

    @parameters.setter
    def parameters(self, value):
        """
        Assigning to this attribute updates the parameters array rather than
        replacing it.
        """

        try:
            value = np.array(value).reshape(self._parameters.shape)
        except ValueError as e:
            raise InputParameterError(
                "Input parameter values not compatible with the model "
                "parameters array: {0}".format(e))

        self._parameters[:] = value

    def __getattr__(self, attr):
        if len(attr) > 1 and attr[0] == '_' and attr != '_param_metrics':
            param_name = attr[1:]
            if param_name in self._param_metrics:
                param_slice, param_shape = self._param_metrics[param_name]
                value = self._parameters[param_slice]
                if self.param_dim == 1:
                    return value[0]
                else:
                    if param_shape:
                        return value.reshape(param_shape)
                    else:
                        return value[0]

        raise AttributeError(attr)

    def __setattr__(self, attr, value):
        if (len(attr) > 1 and attr[0] == '_' and
                hasattr(self, '_param_metrics')):
            param_name = attr[1:]
            if param_name in self._param_metrics:
                # TODO: Maybe handle exception on invalid input shape
                param_slice = self._param_metrics[param_name][0]
                self._parameters[param_slice] = np.array(value).ravel()
                return

        super(ParametricModel, self).__setattr__(attr, value)

    def __repr__(self):
        try:
            degree = str(self.deg)
        except AttributeError:
            degree = ""
        try:
            param_dim = str(self.param_dim)
        except AttributeError:
            param_dim = " "

        if degree:
            fmt = "<{0}({1}, ".format(self.__class__.__name__, repr(self.deg))
        else:
            fmt = "<{0}(".format(self.__class__.__name__)

        for name in self.param_names:
            fmt1 = "{0}={1}, ".format(name, getattr(self, name))
            fmt += fmt1
        if param_dim:
            fmt += "param_dim={0})>".format(self.param_dim)

        return fmt

    def __str__(self):
        try:
            degree = str(self.degree)
        except AttributeError:
            degree = 'N/A'

        fmt = """
        Model: {0}
        n_inputs:   {1}
        Degree: {2}
        Parameter sets: {3}
        Parameters: \n{4}
        """.format(
              self.__class__.__name__,
              self.n_inputs,
              degree,
              self.param_dim,
              indent('\n'.join('{0}: {1}'.format(n, getattr(self, n))
                     for n in self.param_names),
                     width=19))

        return dedent(fmt[1:])

    def set_joint_parameters(self, jparams):
        """
        Used by the `JointFitter` class to store parameters which are
        considered common for several models and are to be fitted together.
        """
        self.joint = jparams

    def _model_to_fit_params(self):
        """
        Create a set of parameters to be fitted.

        These may be a subset of the model parameters, if some of them are held
        constant or tied.
        """
        fitparam_indices = range(len(self.param_names))
        if any(self.fixed.values()) or any(self.tied.values()):
            params = list(self.parameters)
            for idx, name in list(enumerate(self.param_names))[::-1]:
                if self.fixed[name] or self.tied[name]:
                    sl = self._param_metrics[name][0]
                    del params[sl]
                    del fitparam_indices[idx]
            return (np.array(params), fitparam_indices)
        else:
            return (self.parameters, fitparam_indices)

    def _fitter_to_model_params(self, fps):
        _fit_params, _fit_param_indices = self._model_to_fit_params()
        if any(self.fixed.values()) or any(self.tied.values()):
            self.parameters[_fit_param_indices] = fps
            for idx, name in enumerate(self.param_names):
                if self.tied[name] != False:
                    value = self.tied[name](self)
                    slice_ = self._param_metrics[name][0]
                    self.parameters[slice_] = value
        elif any([tuple(b) != (None, None) for b in self.bounds.values()]):
            for name, par in zip(self.param_names, _fit_params):
                if self.bounds[name] != (None, None):
                    b = self.bounds[name]
                    if b[0] is not None:
                        par = max(par, self.bounds[name][0])
                    if b[1] is not None:
                        par = min(par, self.bounds[name][1])
                    setattr(self, name, par)
        else:
            self.parameters = fps

    def _initialize_parameters(self, params):
        """
        Initialize the _parameters array that stores raw parameter values for
        all parameter sets for use with vectorized fitting algorithms; on
        ParametricModels the _param_name attributes actually just reference
        slices of this array.
        """

        # First we need to determine how much array space is needed by all the
        # parameters based on the number of parameters, the shape each input
        # parameter, and the param_dim
        self._param_metrics = {}
        total_size = 0
        for name in self.param_names:

            if params.get(name) is None:
                # parameters that were not supplied at all or that have
                # defaults of None should attempt to use the default provided
                # by their Parameter descriptor
                params[name] = getattr(self, name).default

            value = params[name]

            param_size = np.size(value)
            param_shape = np.shape(value)

            if self.param_dim == 1:
                pass
            elif self.param_dim > 1:
                if param_size == 1:
                    param_size = self.param_dim
                    # Update the value for this param to the new repeated
                    # version
                    value = params[name] = np.repeat(value, param_size)
                    param_shape = value.shape
            else:
                raise ValueError("Model param_dim must be 1 or greater.")

            if (param_size > 1 and param_size != self.param_dim and
                    len(value) != self.param_dim):
                # The len(value) == self.param_dim case is a special case
                # (see #1680) where the parameter has compound values (like [1,
                # 2]) but we're passing in two (or more) param sets, so we want
                # to make sure a value like [[1, 2], [3, 4]] is interpreted as
                # having values for 2 param sets, not 4.

                # For now all parameter values must have a number of elements
                # equal to param_dim (the number of param sets) or it is
                # invalid.  The *only* exception is, again, a scalar value is
                # allowed to be repeated across param sets
                raise InputParameterError(
                    "The input value for {0!r} has too many elements for "
                    "param_dim={1}.  Parameter values must either match the "
                    "model's param_dim in size, or may be a scalar value, in "
                    "which case the value is repeated accross parameter "
                    "sets.".format(name, self.param_dim))

            param_slice = slice(total_size, total_size + param_size)
            self._param_metrics[name] = (param_slice, param_shape)
            total_size += param_size

        self._parameters = np.empty(total_size, dtype=np.float64)

        # Now set the parameter values (this will also fill
        # self._parameters)
        for name, value in params.items():
            setattr(self, name, value)


    def _model_to_fit_params(self):
        """
        Create a set of parameters to be fitted.

        These may be a subset of the model parameters, if some of them are held
        constant or tied.

        This is an adapter of the model parameters to fitters, none of which
        currently support parameters with tied/fixed constraints.

        TODO: It may make more sense to make this adaptation on the Fitter end,
        since one could conceivably implement a fitter from scratch (for
        Astropy specifically) that understands how to use tied constraints, for
        example.
        """

        fitparam_indices = range(len(self.param_names))
        if any(self.fixed.values()) or any(self.tied.values()):
            params = list(self.parameters)
            for idx, name in list(enumerate(self.param_names))[::-1]:
                if self.fixed[name] or self.tied[name]:
                    sl = self._param_metrics[name][0]
                    del params[sl]
                    del fitparam_indices[idx]
            return (np.array(params), fitparam_indices)
        else:
            return (self.parameters, fitparam_indices)


class LabeledInput(dict):
    """
    Create a container with all input data arrays, assigning labels for
    each one.

    Used by CompositeModel to choose input data using labels

    Parameters
    ----------
    data : list
        a list of all input data
    labels : list of strings
        names matching each coordinate in data

    Returns
    -------
    data : LabeledData
        a dict of input data and their assigned labels

    Examples
    --------
    >>> x,y = np.mgrid[:5, :5]
    >>> l = np.arange(10)
    >>> ado = LabeledInput([x, y, l], ['x', 'y', 'pixel'])
    >>> ado.x
    array([[0, 0, 0, 0, 0],
           [1, 1, 1, 1, 1],
           [2, 2, 2, 2, 2],
           [3, 3, 3, 3, 3],
           [4, 4, 4, 4, 4]])
    >>> ado['x']
    array([[0, 0, 0, 0, 0],
           [1, 1, 1, 1, 1],
           [2, 2, 2, 2, 2],
           [3, 3, 3, 3, 3],
           [4, 4, 4, 4, 4]])
    """

    def __init__(self, data, labels):
        dict.__init__(self)
        assert len(labels) == len(data)
        self.labels = [l.strip() for l in labels]
        for coord, label in zip(data, labels):
            self[label] = coord
            setattr(self, '_' + label, coord)
        self._set_properties(self.labels)

    def _getlabel(self, name):
        par = getattr(self, '_' + name)
        return par

    def _setlabel(self, name, val):
        setattr(self, '_' + name, val)
        self[name] = val

    def _dellabel(self, name):
        delattr(self, '_' + name)
        del self[name]

    def add(self, label=None, value=None, **kw):
        """
        Add input data to a LabeledInput object

        Parameters
        --------------
        label : str
            coordinate label
        value : numerical type
            coordinate value
        kw : dictionary
            if given this is a dictionary of {label: value} pairs
        """

        if kw:
            if label is None or value is None:
                self.update(kw)
            else:
                kw[label] = value
                self.update(kw)
        else:
            kw = dict({label: value})
            assert(label is not None and value is not None), (
                "Expected label and value to be defined")
            self[label] = value

        for key in kw:
            self.__setattr__('_' + key, kw[key])
        self._set_properties(kw.keys())

    def _set_properties(self, attributes):
        for attr in attributes:
            setattr(self.__class__, attr, property(lambda self, attr=attr:
                                                   self._getlabel(attr),
                    lambda self, value, attr=attr:
                                                   self._setlabel(attr, value),
                    lambda self, attr=attr:
                                                   self._dellabel(attr)
                                                   )
                    )

    def copy(self):
        data = [self[label] for label in self.labels]
        return LabeledInput(data, self.labels)


class _CompositeModel(Model):
    def __init__(self, transforms, n_inputs, n_outputs):
        """Base class for all composite models."""

        self._transforms = transforms
        param_names = []
        for tr in self._transforms:
            param_names.extend(tr.param_names)
        super(_CompositeModel, self).__init__()
        self.param_names = param_names
        self.n_inputs = n_inputs
        self.n_outputs = n_outputs
        self.fittable = False

    def __repr__(self):
        fmt = """
            Model:  {0}
            """.format(self.__class__.__name__)
        fmt_args = tuple(repr(tr) for tr in self._transforms)
        fmt1 = (" %s  " * len(self._transforms)) % fmt_args
        fmt = fmt + fmt1
        return fmt

    def __str__(self):
        fmt = """
            Model:  {0}
            """.format(self.__class__.__name__)
        fmt_args = tuple(str(tr) for tr in self._transforms)
        fmt1 = (" %s  " * len(self._transforms)) % fmt_args
        fmt = fmt + fmt1
        return fmt

    def add_model(self, transf, inmap, outmap):
        self[transf] = [inmap, outmap]

    def invert(self):
        raise NotImplementedError("Subclasses should implement this")

    def __call__(self):
        # implemented by subclasses
        raise NotImplementedError("Subclasses should implement this")


class SerialCompositeModel(_CompositeModel):
    """
    Composite model that evaluates models in series.

    Parameters
    ----------
    transforms : list
        a list of transforms in the order to be executed
    inmap : list of lists or None
        labels in an input instance of LabeledInput
        if None, the number of input coordinates is exactly what
        the transforms expect
    outmap : list or None
        labels in an input instance of LabeledInput
        if None, the number of output coordinates is exactly what
        the transforms expect

    Notes
    -----
    Output values of one model are used as input values of another.
    Obviously the order of the models matters.

    Examples
    --------
    Apply a 2D rotation followed by a shift in x and y::

        >>> import numpy as np
        >>> from astropy.modeling import models, LabeledInput, SerialCompositeModel
        >>> x, y = np.mgrid[:5, :5]
        >>> rotation = models.MatrixRotation2D(angle=23.5)
        >>> offset_x = models.Shift(-4.23)
        >>> offset_y = models.Shift(2)
        >>> labeled_input = LabeledInput([x, y], ["x", "y"])
        >>> transform = SerialCompositeModel([rotation, offset_x, offset_y],
        ...                                  inmap=[['x', 'y'], ['x'], ['y']],
        ...                                  outmap=[['x', 'y'], ['x'], ['y']])
        >>> result = transform(labeled_input)
    """

    def __init__(self, transforms, inmap=None, outmap=None, n_inputs=None,
                 n_outputs=None):
        if n_inputs is None:
            n_inputs = max([tr.n_inputs for tr in transforms])
            # the output dimension is equal to the output dim of the last
            # transform
            n_outputs = transforms[-1].n_outputs
        else:
            assert n_outputs is not None, "Expected n_inputs and n_outputs"
            n_inputs = n_inputs
            n_outputs = n_outputs

        super(SerialCompositeModel, self).__init__(transforms, n_inputs,
                                                   n_outputs)

        if transforms and inmap and outmap:
            assert len(transforms) == len(inmap) == len(outmap), \
                "Expected sequences of transform, " \
                "inmap and outmap to have the same length"

        if inmap is None:
            inmap = [None] * len(transforms)

        if outmap is None:
            outmap = [None] * len(transforms)

        self._inmap = inmap
        self._outmap = outmap

    def inverse(self):
        try:
            transforms = [tr.inverse() for tr in self._transforms[::-1]]
        except NotImplementedError:
            raise
        if self._inmap is not None:
            inmap = self._inmap[::-1]
            outmap = self._outmap[::-1]
        else:
            inmap = None
            outmap = None
        return SerialCompositeModel(transforms, inmap, outmap)

    def __call__(self, *data):
        """Transforms data using this model."""

        if len(data) == 1:
            if not isinstance(data[0], LabeledInput):
                assert self._transforms[0].n_inputs == 1, \
                    "First transform expects {0} inputs, 1 given".format(
                        self._transforms[0].n_inputs)

                result = data[0]
                for tr in self._transforms:
                    result = tr(result)
                return result
            else:
                labeled_input = data[0].copy()
                # we want to return the entire labeled object because some
                # parts of it may be used in another transform of which this
                # one is a component
                assert self._inmap is not None, \
                    ("Parameter 'inmap' must be provided when "
                     "input is a labeled object.")
                assert self._outmap is not None, \
                    ("Parameter 'outmap' must be provided when input is a "
                     "labeled object")

                for transform, incoo, outcoo in izip(self._transforms,
                                                     self._inmap,
                                                     self._outmap):
                    inlist = [labeled_input[label] for label in incoo]
                    result = transform(*inlist)
                    if len(outcoo) == 1:
                        result = [result]
                    for label, res in zip(outcoo, result):

                        if label not in labeled_input.labels:
                            labeled_input[label] = res
                        setattr(labeled_input, label, res)
                return labeled_input
        else:
            assert self.n_inputs == len(data), \
                "This transform expects {0} inputs".format(self._n_inputs)

            result = self._transforms[0](*data)
            for transform in self._transforms[1:]:
                result = transform(*result)
        return result


class SummedCompositeModel(_CompositeModel):
    """
    Composite model that evaluates models in parallel.

    Parameters
    --------------
    transforms : list
        transforms to be executed in parallel
    inmap : list or None
        labels in an input instance of LabeledInput
        if None, the number of input coordinates is exactly what the
        transforms expect
    outmap : list or None

    Notes
    -----
    Evaluate each model separately and add the results to the input_data.
    """

    def __init__(self, transforms, inmap=None, outmap=None):
        self._transforms = transforms
        n_inputs = self._transforms[0].n_inputs
        n_outputs = n_inputs
        for transform in self._transforms:
            assert transform.n_inputs == transform.n_outputs == n_inputs, \
                ("A SummedCompositeModel expects n_inputs = n_outputs for "
                 "all transforms")

        super(SummedCompositeModel, self).__init__(transforms, n_inputs,
                                                   n_outputs)

        self._inmap = inmap
        self._outmap = outmap

    def __call__(self, *data):
        """Transforms data using this model."""

        if len(data) == 1:
            if not isinstance(data[0], LabeledInput):
                x = data[0]
                deltas = sum(tr(x) for tr in self._transforms)
                return deltas
            else:
                assert self._inmap is not None, \
                    ("Parameter 'inmap' must be provided when "
                     "input is a labeled object.")
                assert self._outmap is not None, \
                    ("Parameter 'outmap' must be provided when input is a "
                     "labeled object")
                labeled_input = data[0].copy()
                # create a list of inputs to be passed to the transforms
                inlist = [getattr(labeled_input, label)
                          for label in self._inmap]
                sum_of_deltas = [np.zeros_like(x) for x in inlist]
                for transform in self._transforms:
                    delta = [transform(*inlist)]
                    for i in range(len(sum_of_deltas)):
                        sum_of_deltas[i] += delta[i]

                for outcoo, delta in izip(self._outmap, sum_of_deltas):
                    setattr(labeled_input, outcoo, delta)
                # always return the entire labeled object, not just the result
                # since this may be part of another composite transform
                return labeled_input
        else:
            result = self._transforms[0](*data)
            assert self.n_inputs == self.n_outputs
            for tr in self._transforms[1:]:
                result += tr(*data)
            return result


class Parametric1DModel(ParametricModel):
    """
    Base class for one dimensional parametric models.

    This class provides an easier interface to defining new models.
    Examples can be found in functional_models.py

    Parameters
    ----------
    parameters : dictionary
        Dictionary of model parameters with initialisation values
        {'parameter_name': 'parameter_value'}
    """

    @format_input
    def __call__(self, x):
        """
        Transforms data using this model.

        Parameters
        ----------
        x : array like or a number
            input
        """

        return self.eval(x, *self.param_sets)


class Parametric2DModel(ParametricModel):
    """
    Base class for two dimensional parametric models.

    This class provides an easier interface to defining new models.
    Examples can be found in functional_models.py

    Parameters
    ----------
    parameter_dict : dictionary
        Dictionary of model parameters with initialization values
        {'parameter_name': 'parameter_value'}
    """

    n_inputs = 2
    n_outputs = 1

    @format_input
    def __call__(self, x, y):
        """
        Transforms data using this model.

        Parameters
        ----------
        x : array like or a number
            input
        """

        return self.eval(x, y, *self.param_sets)
